#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sun May  5 22:03:32 2019

@author: jkl
"""
EMagPy_version = '0.0.4'
name = 'emagpy'
from emagpy.Problem import Problem
from emagpy.Survey import Survey



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Apr 16 20:29:19 2019

@author: jkl
"""
import os
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from matplotlib.collections import PolyCollection
from scipy.optimize import minimize
from scipy.stats import linregress

from emagpy.invertHelper import (fCS, fMaxwellECa, fMaxwellQ, buildSecondDiff,
                                 buildJacobian, getQs, eca2Q, Q2eca2, Q2eca)
from emagpy.Survey import Survey, idw, clipConvexHull, griddata


class Problem(object):
    """Class defining an inversion problem.
    
    """
    def __init__(self):
        self.depths0 = np.array([0.5, 1.5]) # initial depths of the bottom of each layer (last one is -inf)
        self.conds0 = np.array([20.0, 20.0, 20.0]) # initial conductivity for each layer
        self.fixedConds = np.array([False, False, False])
        self.fixedDepths = np.array([True, True])
        self.surveys = []
        self.models = [] # contains conds TODO rename to conds ?
        self.rmses = []
        self.freqs = []
        self.depths = [] # contains inverted depths or just depths0 if fixed
        self.ikill = False # if True, the inversion is killed
        
        
    def createSurvey(self, fname, freq=None, hx=None):
        """Create a survey object.
        
        Parameters
        ----------
        fname : str
            Path to the csv file with the data.
        freq : float, optional
            Frequency for all the coils (can also be specified for each coil in the file).
        hx : float, optional
            Height of the instrument above the ground (can also be specified for each coil in the file).
        """
        survey = Survey(fname, freq=freq, hx=hx)
        if len(self.surveys) == 0:
            self.coils = survey.coils
            self.freqs = survey.freqs
            self.cspacing = survey.cspacing
            self.cpos = survey.cpos
            self.hx = survey.hx
            self.surveys.append(survey)
        else: # check we have the same configuration
            check = [a == b for a,b, in zip(self.coils, survey.coils)]
            if all(check) is True:
                self.surveys.append(survey)
        
        
    def createTimeLapseSurvey(self, dirname):
        """ Create a list of surveys object.
        
        Parameters
        ----------
        dirname : str
            Directory with files to be parsed or list of file to be parsed.
        """
        files = dirname
        if isinstance(dirname, list): # it's a list of filename
            if len(dirname) < 2:
                raise ValueError('at least two files needed for timelapse inversion')
            files = dirname
        else: # it's a directory and we import all the files inside
            if os.path.isdir(dirname):
                files = [os.path.join(dirname, f) for f in np.sort(os.listdir(dirname)) if f[0] != '.']
                # this filter out hidden file as well
            else:
                raise ValueError('dirname should be a directory path or a list of filenames')
        for f in files:
            self.createSurvey(f)
        
    
    def importGF(self, fnameLo=None, fnameHi=None, device='CMD Mini-Explorer', hx=0):
        """Import GF instrument data with Lo and Hi file mode. If spatial data
        a regridding will be performed to match the data.
        
        Parameters
        ----------
        fnameLo : str
            Name of the file with the Lo settings.
        fnameHi : str
            Name of the file with the Hi settings.
        device : str, optional
            Type of device. Default is Mini-Explorer.
        hx : float, optional
            Height of the device above the ground in meters according to the
            calibration use (e.g. `F-Ground` -> 0 m, `F-1m` -> 1 m).
        """
        survey = Survey()
        survey.importGF(fnameLo, fnameHi, device, hx)
        self.coils = survey.coils
        self.freqs = survey.freqs
        self.cspacing = survey.cspacing
        self.cpos = survey.cpos
        self.hx = survey.hx
        self.surveys.append(survey)
        
        
    def setDepths(self, depths):
        """ Set the depths of the bottom of each layer. Last layer goes to -inf.
        Depths should be positive going down.
        """
        if len(depths) == 0:
            raise ValueError('No depths specified.')
        if all(np.diff(depths) > 0):
            raise ValueError('Depths should be ordered and increasing.')
        self.depths = np.array(depths)
        
    
        
    def invert(self, forwardModel='CS', method='L-BFGS-B', regularization='l2',
               alpha=0.07, beta=0.0, gamma=0.0, dump=None, bnds=None,
               options={}, Lscaling=False, rep=100):
        """Invert the apparent conductivity measurements.
        
        Parameters
        ----------
        forwardModel : str, optional
            Type of forward model:
                - CS : Cumulative sensitivity (default)
                - FS : Full Maxwell solution with low-induction number (LIN) approximation
                - FSandrade : Full Maxwell solution without LIN approximation (see Andrade 2016)
                - CSgn : Cumulative sensitivity with jacobian matrix (using Gauss-Newton)
                - CSgndiff : Cumulative sensitivty for difference inversion - NOT IMPLEMENTED YET
        method : str, optional
            Name of the optimization method either L-BFGS-B, TNC, CG or Nelder-Mead
            to be passed to `scipy.optimize.minmize()` or ROPE, SCEUA, DREAM for
            a MCMC-based solver based on the `spotpy` Python package.
        regularization : str, optional
            Type of regularization, either l1 (blocky model) or l2 (smooth model)
        alpha : float, optional
            Smoothing factor for the inversion.
        beta : float, optional
            Smoothing factor for neightbouring profile.
        gamma : float, optional
            Smoothing factor between surveys (for time-lapse only).
        dump : function, optional
            Function to print the progression. Default is `print`.
        bnds : list of float, optional
            If specified, will create bounds for the inversion. Doesn't work with
            Nelder-Mead solver.
        options : dict, optional
            Additional dictionary arguments will be passed to `scipy.optimize.minimize()`.
        Lscaling : bool, optional
            If True the regularization matrix will be weighted based on 
            centroids of layers differences.
        rep : int, optional
            Number of sample for the MCMC-based methods.
        """
        # switch in case Gauss-Newton software is selected
        if forwardModel in ['CSgn', 'CSgndiff']:
            self.invertGN(alpha=alpha, alpha_ref=None, dump=dump)
            return
        
        self.models = []
        self.depths = []
        self.rmses = []
        self.ikill = False
        mMinimize = ['L-BFGS-B','TNC','CG','Nelder-Mead']
        mMCMC = ['ROPE','SCEUA','DREAM']
        
        if dump is None:
            def dump(x):
                print('\r' + x, end='')

        nc = len(self.conds0)
        vd = ~self.fixedDepths # variable depths
        vc = ~self.fixedConds # variable conductivity

        # define bounds
        if bnds is not None:
            if len(bnds) == 2: # we just have min/max of EC
                top = np.ones(nc)*bnds[1]
                bot = np.ones(nc)*bnds[0]
                bounds = list(tuple(zip(bot[vc], top[vc])))
            else:
                bounds = bnds
        else:
            bounds = None
        if ((np.sum(vd) > 0) or (method in mMCMC)) and (bounds is None):
            mdepths = self.depths0[:-1] + np.diff(self.depths0)/2
            bot = np.r_[np.r_[0.2, mdepths], np.ones(nc)*2]
            top = np.r_[np.r_[mdepths, self.depths0[-1] + 0.2], np.ones(nc)*100]
            bounds = list(tuple(zip(bot[np.r_[vd, vc]], top[np.r_[vd, vc]])))
        dump('bounds = ' + str(bounds))
#            beta = 0 # not lateral smoothing with changing depths
        
        # define the forward model
        def fmodel(p): # p contains first the depths then the conductivities
            depth = self.depths0
            if np.sum(vd) > 0:
                depth[vd] = p[:np.sum(vd)]
            cond = self.conds0
            if np.sum(vc) > 0:
                cond[vc] = p[np.sum(vd):]
            if forwardModel == 'CS':
                return fCS(cond, depth, self.cspacing, self.cpos, hx=self.hx[0])
            elif forwardModel == 'FS':
                return fMaxwellECa(cond, depth, self.cspacing, self.cpos, f=self.freqs[0], hx=self.hx[0])
            elif forwardModel == 'FSandrade':
                return fMaxwellQ(cond, depth, self.cspacing, self.cpos, f=self.freqs[0], hx=self.hx[0])
            elif forwardModel == 'Q':
                return np.imag(getQs(cond, depth, self.cspacing, self.cpos, f=self.freqs[0], hx=self.hx[0]))


        # build roughness matrix
        L = buildSecondDiff(len(self.conds0)) # L is used inside the smooth objective fct
        # each constrain is proportional to the distance between the centroid of the two layers
        if len(self.depths0) > 1:
            centroids = np.r_[self.depths0[0]/2, self.depths0[:-1] + np.diff(self.depths0)/2]
            if len(self.depths0) > 2:
                distCentroids = np.r_[centroids[1] - centroids[0],
                                     centroids[2:] - centroids[:-2],
                                     centroids[-1] - centroids[-2],
                                     1] # last layer is infinite so we don't apply any weights
            else:
                distCentroids = np.r_[centroids[1] - centroids[0],
                                      centroids[-1] - centroids[-2],
                                      1]
            if Lscaling is True:
                L = L/distCentroids[:,None]
        # TODO what for 3 layers ? sum of those distances ?

        # data misfit
        def dataMisfit(p, obs):
            misfit = fmodel(p) - obs
            if forwardModel == 'Q':
                misfit = misfit * 1e5 # to help the solver with small Q
            return misfit 
        
        # model misfit only for conductivities not depths
        def modelMisfit(p):
            cond = self.conds0
            if np.sum(vc) > 0:
                cond[vc] = p[:np.sum(vc)]
            return np.dot(L, cond)
        
        # set up regularisation
        if regularization  == 'l1':
            def objfunc(p, app, pn):
                return np.sqrt(np.sum(np.abs(dataMisfit(p, app)))/len(app)
                               + alpha*np.sum(np.abs(modelMisfit(p)))/nc
                               + beta*np.sum(np.abs(p - pn))/len(p))
        elif regularization == 'l2':
            def objfunc(p, app, pn):
                return np.sqrt(np.sum(dataMisfit(p, app)**2)/len(app)
                               + alpha*np.sum(modelMisfit(p)**2)/nc
                               + beta*np.sum((p - pn)**2)/len(p))
            
        # define spotpy class if MCMC-based methods
        if method in mMCMC:
            try:
                import spotpy
            except ImportError:
                print('Please install spotpy to use `invertSCUEA()` method.')
                return
            
            class spotpy_setup(object):
                def __init__(self, obsVals, fmodel, bounds):
                    self.params = []
                    for i, bnd in enumerate(bounds):
                        self.params.append(
                                spotpy.parameter.Uniform(
                                        'x{:d}'.format(i), bnd[0], bnd[1], 10, 10, bnd[0], bnd[1]))
                    self.obsVals = obsVals
                    self.fmodel = fmodel
                    
                def parameters(self):
                    return spotpy.parameter.generate(self.params)
            
                def simulation(self, vector):
                    x = np.array(vector)
                    simulations = self.fmodel(x).flatten()
                    return simulations
                
                def evaluation(self): # what the function return when called with the optimal values
                    observations = self.obsVals.flatten()
                    return observations
                
                def objectivefunction(self, simulation, evaluation):
                    objfunc = -spotpy.objectivefunctions.rmse(evaluation, simulation)
                    return objfunc
            
        # inversion row by row
        c = 0 # number of inversion that converged
        x0 = np.r_[self.depths0[vd], self.conds0[vc]]
        for i, survey in enumerate(self.surveys):
            if self.ikill:
                break
            c = 0
            apps = survey.df[self.coils].values
            rmse = np.zeros(apps.shape[0])*np.nan
            model = np.ones((apps.shape[0], len(self.conds0)))*self.conds0
            depth = np.ones((apps.shape[0], len(self.depths0)))*self.depths0
            dump('Survey {:d}/{:d}\n'.format(i+1, len(self.surveys)))
            for j in range(survey.df.shape[0]): # this could be done in //
                if self.ikill:
                    break
                
                # define observations and convert to Q if needed
                obs = apps[j,:]
                if forwardModel == 'Q':
                    obs = np.array([eca2Q(a*1e-3, s) for a, s in zip(obs, self.cspacing)])
                
                # define previous profile in case we want to lateral constrain
                if j == 0:
                    pn = np.zeros(np.sum(np.r_[vd, vc]))
                else:
                    pn = np.r_[depth[j-1,:][vd], model[j-1,:][vc]]

                # solve optimization problem
                if method in mMinimize: # minimize
                    res = minimize(objfunc, x0, args=(obs, pn),
                                   method=method, bounds=bounds, options=options)
                    out = res.x
                    status = 'converged' if res.success else 'not converged'
                    if res.success:
                        c += 1
                elif method in mMCMC: # MCMC based methods
                    cols = ['parx{:d}'.format(a) for a in range(len(bounds))]
                    spotpySetup = spotpy_setup(obs, fmodel, bounds)
                    if method == 'ROPE':
                        sampler = spotpy.algorithms.rope(spotpySetup, dbname='db', dbformat='csv')
                    elif method == 'DREAM':
                        sampler = spotpy.algorithms.dream(spotpySetup, dbname='db', dbformat='csv')
                    elif method == 'SCEUA':
                        sampler = spotpy.algorithms.sceua(spotpySetup, dbname='db', dbformat='csv')
                    else:
                        raise ValueError('Method {:s} unkown'.format(method))
                        return
                    sampler.sample(rep)
                    results = np.array(sampler.getdata())
                    ibest = np.argmin(np.abs(results['like1']))
                    out = np.array(list(results[ibest][cols]))
                    status = 'ok'
                    c += 1

                # store results from optimization
                depth[j,vd] = out[:np.sum(vd)]
                model[j,vc] = out[np.sum(vd):]
                rmse[j] = np.sqrt(np.sum(dataMisfit(out, obs)**2)/len(obs))
                dump('{:d}/{:d} inverted ({:s})'.format(j+1, apps.shape[0], status))
            self.models.append(model)
            self.depths.append(depth)
            self.rmses.append(rmse)
            dump('{:d} measurements inverted ({:d} converged)\n'.format(apps.shape[0], c))
                    
    # TODO add smoothing 3D: maybe invert all profiles once with GN and then
    # invert them again with a constrain on the 5 nearest profiles by distance

    
    
    def invertGN(self, alpha=0.07, alpha_ref=None, dump=None):
        """Fast inversion usign Gauss-Newton and cumulative sensitivity.
        
        Parameters
        ----------
        alpha : float, optional
            Smoothing factor.
        alpha_ref : float, optional
            Only used for difference inversion to contrain the bottom of
            the profile to not changing (see Annex in Whalley et al., 2017).
        dump : function, optional
            Function to output the running inversion.
        """
        self.models = []
        self.rmses = []
        self.depths = []
        
        if dump is None:
            def dump(x):
                print('\r' + x, end='')
                
        J = buildJacobian(self.depths0, self.cspacing, self.cpos)
        L = buildSecondDiff(J.shape[1])
        def fmodel(p):
            return fCS(p, self.depths0, self.cspacing, self.cpos, hx=self.hx[0])
        
        # fCS is automatically adding a leading 0 but not buildJacobian
        def dataMisfit(p, app):
            return app - fmodel(p)
        def modelMisfit(p):
            return np.dot(L, p)
        
        for i, survey in enumerate(self.surveys):
            apps = survey.df[self.coils].values
            rmse = np.zeros(apps.shape[0])*np.nan
            model = np.zeros((apps.shape[0], len(self.conds0)))*np.nan
            dump('Survey {:d}/{:d}\n'.format(i+1, len(self.surveys)))
            for j in range(survey.df.shape[0]):
                app = apps[j,:]
                cond = np.ones((len(self.conds0),1))*np.nanmean(app) # initial EC is the mean of the apparent (doesn't matter)
                # OR search for best starting model here
                for l in range(1): # FIXME this is diverging with time ..;
                    d = dataMisfit(cond, app)
                    LHS = np.dot(J.T, J) + alpha*L
                    RHS = np.dot(J.T, d[:,None]) - alpha*np.dot(L, cond) # minux or plus doesn't matter here ?!
                    if alpha_ref is not None: # constraint the change of the last element of the profile
                        LHS[-1:,-1:] = alpha_ref
                        RHS[-1:] = alpha_ref*cond[i,-1]
                    solution = np.linalg.solve(LHS, RHS)
                    cond = cond + solution # it's an iterative process but it converges in one iteration as it's linear
                out = cond.flatten()
                model[j,:] = out
                rmse[j] = np.sqrt(np.sum(dataMisfit(out, app)**2)/len(app))
                dump('{:d}/{:d} inverted'.format(j+1, apps.shape[0]))
            self.models.append(model)
            self.rmses.append(rmse)
            depth = np.repeat(self.depths0[None,:], apps.shape[0], axis=0)
            self.depths.append(depth)
            dump('{:d} measurements inverted\n'.format(apps.shape[0]))
           


    def tcorrECa(self, tdepths, tprofile):
        """Temperature correction using XXXX formula.
        
        Parameters
        ----------
        tdepths : list of arrays
            Depths in meters of the temperature sensors (negative downards).
        tprofile : list of arrays
            Temperature values corresponding in degree Celsius.
        """
        for i, s in enumerate(self.surveys):
            s.tcorr(tdepths[i], tprofile[i])


            
    def tcorrEC(self, tdepths, tprofile):
        """Temperature correction for inverted models using XXXX formula.
        
        Parameters
        ----------
        tdepths : array-like
            Depths in meters of the temperature sensors (negative downards).
        tprofile : array-like
            Temperature values corresponding in degree Celsius.
        """
        for i, model in enumerate(self.models):
            pass
        #TODO correct ECa or inverted EC ? maybe let this to the user


    def write2vtk(self):
        """Write .vtk cloud points with the inverted models.
        """
        for i, m in enumerate(self.models):
            
            pass


    def rollingMean(self, window=3):
        """Perform a rolling mean on the data.
        
        Parameters
        ----------
        window : int, optional
            Size of the windows for rolling mean.
        """
        for survey in self.surveys:
            survey.rollingMean(window=window)
            
    
    
    def filterDiff(self, coil=None, thresh=5):
        """Filter out consecutive measurements when the difference between them
        is larger than val.
        
        Parameters
        ----------
        thresh : float, optional
            Value of absolute consecutive difference above which the second 
            data point will be discarded.
        coil : str, optional
            Coil on which to apply the processing.
        """
        for s in self.surveys:
            s.filterDiff(coil=coil, thresh=thresh)
            
            
        
    def forward(self, forwardModel='CS', coils=None, noise=0.0):
        """Forward model.
        
        Parameters
        ----------
        forwardModel : str, optional
            Type of forward model:
                - CS : Cumulative sensitivity (default)
                - FS : Full Maxwell solution with low-induction number (LIN) approximation
                - FSandrade : Full Maxwell solution without LIN approximation (see Andrade 2016)
                - CSfast : Cumulative sensitivity with jacobian matrix (not minimize) - NOT IMPLEMENTED YET
                - CSdiff : Cumulative sensitivty for difference inversion - NOT IMPLEMENTED YET
        coils : list of str, optional
            If `None`, then the default attribute of the object will be used (foward
            mode on inverted solution).
            If specified, the coil spacing, orientation and height above the ground
            will be set. In this case you need to assign at models and depths (full forward mode).
            The ECa values generated will be incorporated as a new Survey object.
        noise : float, optional
            Percentage of noise to add on the generated apparent conductivities.
        
        Returns
        -------
        df : pandas.DataFrame
            With the apparent ECa in the same format as input for the Survey class.
        If `coils` argument is specified a new Survey object will be added as well.
        """
        if coils is None: # forward mode on inverted solution
            cspacing = self.cspacing
            cpos = self.cpos
            hxs = self.hx
            freqs = self.freqs
            iForward = False
        else: # full forward mode
            print('Forward modelling')
            iForward = True
            self.coils = coils
            self.depths0 = self.depths[0][0,:]
            self.conds0 = self.models[0][0,:]
            cspacing = []
            cpos = []
            hxs = []
            freqs = []
            for arg in coils:
                cpos.append(arg[:3].lower())
                b = arg[3:].split('f')
                cspacing.append(float(b[0]))
                if len(b) > 1:
                    c = b[1].split('h')
                    freqs.append(float(c[0]))
                    if len(c) > 1:
                        hxs.append(float(c[1]))
                    else:
                        hxs.append(0)
                else:
                    freqs.append(30000) # Hz default is not specified !!
                    hxs.append(0)
            self.cspacing = cspacing
            self.cpos = cpos
            self.hx = hxs
            self.freqs = freqs
            
        
        if forwardModel in ['CS','FS','FSandrade']:
            # define the forward model
            if forwardModel == 'CS':
                def fmodel(p, depth):
                    return fCS(p, depth, cspacing, cpos, hx=hxs[0])
            elif forwardModel == 'FS':
                def fmodel(p, depth):
                    return fMaxwellECa(p, depth, cspacing, cpos, f=freqs[0], hx=hxs[0])
            elif forwardModel == 'FSandrade':
                def fmodel(p, depth):
                    return fMaxwellQ(p, depth, cspacing, cpos, f=freqs[0], hx=hxs[0])
        
        def addnoise(x, level=0.05):
            return x + np.random.randn(len(x))*x*level
        
        dfs = []
        for i, model in enumerate(self.models):
            depths = self.depths[i]
            apps = np.zeros((model.shape[0], len(self.coils)))*np.nan
            for j in range(model.shape[0]):
                conds = model[j,:]
                depth = depths[j,:]
                apps[j,:] = addnoise(fmodel(conds, depth), level=noise)
        
            df = pd.DataFrame(apps, columns=self.coils)
            dfs.append(df)
        
        if iForward:
            self.surveys = []
            for df in dfs:
                s = Survey()
                s.readDF(df)
                self.surveys.append(s)
        
        return dfs
    
    
    
    def show(self, index=0, **kwargs):
        """Show the raw data of the survey.
        
        Parameters
        ----------
        index : int, optional
            Survey number, by default, the first survey is chosen.
        """
        coil = kwargs['coil'] if 'coil' in kwargs else 'all'
        vmin = kwargs['vmin'] if 'vmin' in kwargs else None
        vmax = kwargs['vmax'] if 'vmax' in kwargs else None
        ax = kwargs['ax'] if 'ax' in kwargs else None
        self.surveys[index].show(coil=coil, vmin=vmin, vmax=vmax, ax=ax)
    
    
    def showMap(self, index=0, **kwargs):
        """Show spatial map of the selected survey.
        
        Parameters
        ----------
        index : int, optional
            Survey number, by default, the first survey is chosen.
        """
        coil = kwargs['coil'] if 'coil' in kwargs else None
        vmin = kwargs['vmin'] if 'vmin' in kwargs else None
        vmax = kwargs['vmax'] if 'vmax' in kwargs else None
        contour = kwargs['contour'] if 'contour' in kwargs else False
        pts = kwargs['pts'] if 'pts' in kwargs else False
        cmap = kwargs['cmap'] if 'cmap' in kwargs else 'viridis_r'
        ax = kwargs['ax'] if 'ax' in kwargs else None
        self.surveys[index].showMap(coil=coil, vmin=vmin, vmax=vmax,
                    contour=contour, pts=pts, cmap=cmap, ax=ax)
        
    
    def saveMap(self, index=0, **kwargs):
        """Save georefenced .tiff.
        
        Parameters
        ----------
        index : int, optional
            Survey number, by default, the first survey is chosen.
        """
        self.surveys[index].saveMap(**kwargs)
    
    
    def saveSlice(self, fname, index=0, islice=0, nx=100, ny=100, method='nearest',
                xmin=None, xmax=None, ymin=None, ymax=None, color=False,
                cmap='viridis', vmin=None, vmax=None):
        """Save a georeferenced raster TIFF file for the specified inverted depths.
        
        Parameters
        ----------
        fname : str
            Path of where to save the .tiff file.
        index : int, optional
            Survey index. Default is first.
        islice : int, optional
            Depth index. Default is first depth.
        nx : int, optional
            Number of points in x direction.
        ny : int, optional
            Number of points in y direction.
        method : str, optional
            Interpolation method (nearest, cubic or linear see
            `scipy.interpolate.griddata`) or IDW (default).
        xmin : float, optional
            Mininum X value.
        xmax : float, optional
            Maximum X value.
        ymin : float, optional
            Minimium Y value.
        ymax : float, optional
            Maximum Y value
        color : bool, optional
            If True a colormap will be applied.
        cmap : str, optional
            If `color == True`, name of the colormap. Default is viridis.
        vmin : float, optional
            Minimum value for colomap.
        vmax : float, optional
            Maximum value for colormap.
        """
        import rasterio
        from rasterio.transform import from_origin

        values = self.models[index][:,islice]
        xknown = self.surveys[index].df['x'].values
        yknown = self.surveys[index].df['y'].values

        if xmin is None:
            xmin = np.min(xknown)
        if xmax is None:
            xmax = np.max(xknown)
        if ymin is None:
            ymin = np.min(yknown)
        if ymax is None:
            ymax = np.max(yknown)
        X, Y = np.meshgrid(np.linspace(xmin, xmax, nx),
                           np.linspace(ymin, ymax, ny))
        x, y = X.flatten(), Y.flatten()
        if method == 'idw':
            z = idw(x, y, xknown, yknown, values)
            z = z.reshape(X.shape)
        else:
            z = griddata(np.c_[xknown, yknown], values, (X, Y), method=method)
        inside = np.ones(nx*ny)
        inside2 = clipConvexHull(xknown, yknown, x, y, inside)
        ie = np.isnan(inside2).reshape(z.shape)
        z[ie] = np.nan
        Z = np.flipud(z.T)
        
        # distance between corners
        dist0 = np.abs(xmax - xmin)
        dist1 = np.abs(ymax - ymin)        
    
        Z = np.fliplr(np.flipud(Z.T))
        yscale = dist1/Z.shape[0]
        xscale = dist0/Z.shape[1]
        
        tOffsetScaling = from_origin(xmin - xscale/2, ymax - yscale/2, xscale, yscale)
        tt = tOffsetScaling
        
        if color == True:
            if vmin is None:
                vmin = np.nanpercentile(Z.flatten(), 2)
            if vmax is None:
                vmax = np.nanpercentile(Z.flatten(), 98)
            norm = plt.Normalize(vmin=vmin, vmax=vmax)
            Z = plt.get_cmap(cmap)(norm(Z))
            Z = 255*Z
            Z = Z.astype('uint8')
            for i in range(4):
                Z[np.fliplr(ie.T).T, i] = 0
        
            with rasterio.open(fname, 'w',
                           driver='GTiff',
                           height=Z.shape[0],
                           width=Z.shape[1], count=4, dtype=Z.dtype,
                           crs='+init=epsg:27700', transform=tt) as dst:
                for i in range(4):
                    dst.write(Z[:,:,i], i+1)
        else:
            with rasterio.open(fname, 'w',
                               driver='GTiff',
                               height=Z.shape[0],
                               width=Z.shape[1], count=1, dtype=Z.dtype,
                               crs='+init=epsg:27700', transform=tt) as dst:
                dst.write(Z, 1)
        
            
    def gridData(self, nx=100, ny=100, method='nearest'):
        """ Grid data (for 3D).
        
        Parameters
        ----------
        nx : int, optional
            Number of points in x direction.
        ny : int, optional
            Number of points in y direction.
        method : str, optional
            Interpolation method (nearest, cubic or linear see
            `scipy.interpolate.griddata`). Default is `nearest`.
        """
        
        for survey in self.surveys:
            survey.gridData(nx=nx, ny=ny, method=method)
        
    
    def setInit(self, depths0, conds0=None, fixedDepths=None, fixedConds=None):
        """Set the initial depths and conductivity for the inversion.
        
        Parameters
        ----------
        depths0 : list or array
            Depth as positive number of the bottom of each layer.
            There is N-1 depths for N layers as the last layer is infinite.
        conds0 : list or array, optional
            Starting conductivity in mS/m of each layer.
            By default a homogeneous conductivity of 20 mS/m is defined.
        fixedDepths : array of bool, optional
            Boolean array of same length as `depths0`. True if depth is fixed.
            False if it's a parameter. By default all depths are fixed.
        fixedConds : array of bool, optional
            Boolean array of same length as `conds0`. True if conductivity if fixed.
            False if it's a parameter. By default all conductivity are variable.'
        """
        depths0 = np.array(depths0)
        if np.sum(depths0 < 0) > 0:
            raise ValueError('All depth should be specified as positive number.')
        if np.sum(depths0 == 0) > 0:
            raise ValueError('No depth should be equals to 0 (infinitely thin layer)')
        if conds0 is None:
            conds0 = np.ones(len(depths0)+1)*20
        if fixedDepths is None:
            fixedDepths = np.ones(len(depths0), dtype=bool)
        if fixedConds is None:
            fixedConds = np.zeros(len(conds0), dtype=bool)
        if len(fixedConds) != len(conds0):
            raise ValueError('len(fixedConds) should be equal to len(conds0).')
        if len(fixedDepths) != len(depths0):
            raise ValueError('len(fixedDepths) should be equal to len(depths0)')    
        if len(depths0) + 1 != len(conds0):
            raise ValueError('length of conds0 should be equals to length of depths0 + 1')
        else:
            self.depths0 = depths0
            self.conds0 = np.array(conds0, dtype=float)
            self.fixedDepths = np.array(fixedDepths, dtype=bool)
            self.fixedConds = np.array(fixedConds, dtype=bool)


    
    def convertFromNMEA(self,  targetProjection='EPSG:27700'): # British Grid 1936
        """ Convert NMEA string to selected CRS projection.
        
        Parameters
        ----------
        targetProjection : str, optional
            Target CRS, in EPSG number: e.g. `targetProjection='EPSG:27700'`
            for the British Grid.
        """
        for survey in self.surveys:
            survey.convertFromNMEA(targetProjection=targetProjection)
    
    
    
    def showResults_old(self, index=0, ax=None, vmin=None, vmax=None,
                    maxDepth=None, padding=1, cmap='viridis_r'):
        """Show invertd model.
        
        Parameters
        ----------
        index : int, optional
            Index of the survey to plot.
        ax : Matplotlib.Axes, optional
            If specified, the graph will be plotted against this axis.
        vmin : float, optional
            Minimum value of the colorbar.
        vmax : float, optional
            Maximum value of the colorbar.
        maxDepth : float, optional
            Maximum negative depths of the graph.
        padding : float, optional
            Thickness of the bottom infinite layer in [m].
        cmap : str, optional
            Name of the Matplotlib colormap to use.
        """            
        sig = self.models[index]
        x = np.arange(sig.shape[0])
#        depths = np.repeat(self.depths0[:,None], sig.shape[0], axis=1).T
        depths = self.depths[0]      
        
        if depths[0,0] != 0:
            depths = np.c_[np.zeros(depths.shape[0]), depths]
        if vmin is None:
            vmin = np.nanpercentile(sig, 5)
        if vmax is None:
            vmax = np.nanpercentile(sig, 95)
        cmap = plt.get_cmap(cmap)
        norm = plt.Normalize(vmin=vmin, vmax=vmax)
        if maxDepth is None:
            maxDepth = np.max(depths) + padding
        depths = np.c_[depths, np.ones(depths.shape[0])*maxDepth]
        h = np.diff(depths, axis=1)
        h = np.c_[np.zeros(h.shape[0]), h]
        if ax is None:
            fig, ax = plt.subplots()
        else:
            fig = ax.figure
         # optimization doesn't seem to work maybe it needs only increasing arrays...   
#        ax.bar(np.tile(x, h.shape[1]-1), -h[:,1:].flatten('F'), bottom=-np.cumsum(h, axis=1)[:,1:].flatten('F'),
#               color=cmap(norm(sig.flatten('F'))), edgecolor='none', width=1)
        for i in range(1, h.shape[1]):
            ax.bar(x, -h[:,i], bottom=-np.sum(h[:,:i], axis=1),
                   color=cmap(norm(sig[:,i-1])), edgecolor='none', width=1)
        sm = plt.cm.ScalarMappable(cmap=cmap, norm=norm)
        sm.set_array([])
        fig.colorbar(sm, label='Conductivity [mS/m]')
        ax.set_xlabel('X position')
        ax.set_ylabel('Depth [m]')
        ax.set_title(self.surveys[index].name)
        ax.set_ylim([-maxDepth, 0])
#        ax.set_aspect('equal')
        def format_coord(i,j):
            col=int(np.floor(i))+1
            if col < sig.shape[0]:
                row = int(np.where(-depths[col,:] < j)[0].min())-1
                return 'x={0:.4f}, y={1:.4f}, value={2:.4f}'.format(col, row, sig[col, row])
            else:
                return ''
        ax.format_coord = format_coord
        fig.tight_layout()


    
    def showResults(self, index=0, ax=None, vmin=None, vmax=None,
                    maxDepth=None, padding=1, cmap='viridis_r',
                    contour=False):
        """Show inverted model.
        
        Parameters
        ----------
        index : int, optional
            Index of the survey to plot.
        ax : Matplotlib.Axes, optional
            If specified, the graph will be plotted against this axis.
        vmin : float, optional
            Minimum value of the colorbar.
        vmax : float, optional
            Maximum value of the colorbar.
        maxDepth : float, optional
            Maximum negative depths of the graph.
        padding : float, optional
            DONT'T KNOW
        cmap : str, optional
            Name of the Matplotlib colormap to use.
        contour : bool, optional
            If `True` a contour plot will be plotted.
        """            
        sig = self.models[index]
        x = np.arange(sig.shape[0])
#        x = np.sqrt(np.diff(self.surveys[index].df[['x', 'y']].values, axis=1)**2)
#        depths = np.repeat(self.depths0[:,None], sig.shape[0], axis=1).T
        depths = self.depths[0]
        
        if ax is None:
            fig, ax = plt.subplots()
        else:
            fig = ax.figure
        if depths[0,0] != 0:
            depths = np.c_[np.zeros(depths.shape[0]), depths]
        if vmin is None:
            vmin = np.nanpercentile(sig, 5)
        if vmax is None:
            vmax = np.nanpercentile(sig, 95)
        cmap = plt.get_cmap(cmap)
        if maxDepth is None:
            maxDepth = np.max(depths) + padding
        depths = np.c_[depths, np.ones(depths.shape[0])*maxDepth]
        
        # vertices
        nlayer = sig.shape[1]
        nsample = sig.shape[0]
        x2 = np.arange(nsample+1)
#        dist = np.sqrt((self.surveys[index].df['x'].values - self.surveys[index].df['y'].values)**2)
        xs = np.tile(np.repeat(x2, 2)[1:-1][:,None], nlayer+1)
        ys = -np.repeat(depths, 2, axis=0)
        vertices = np.c_[xs.flatten('F'), ys.flatten('F')]
        
        # connection matrix
        n = vertices.shape[0]
        connection = np.c_[np.arange(n).reshape(-1,2),
                           2*nsample + np.arange(n).reshape(-1,2)[:,::-1]]
#        ie1 = connection[:,0] % (2*len(x)-2) == 0
        ie2 = (connection >= len(vertices)).any(1)
        ie = ie2
        connection = connection[~ie, :]
        coordinates = vertices[connection]
        
        # plotting
        if contour is True:
            centroid = np.mean(coordinates, axis=1)
            x = np.r_[centroid[:nsample,0], centroid[:,0], centroid[:nsample,0]]
            y = np.r_[np.zeros(nsample), centroid[:,1], -np.ones(nsample)*maxDepth]
            z = np.c_[sig[:,0], sig, sig[:,-1]]
            if vmax > vmin:
                levels = np.linspace(vmin, vmax, 7)
            else:
                levels = None
            cax = ax.tricontourf(x, y, z.flatten('F'),
                                 cmap=cmap, levels=levels, extend='both')
#            ax.plot(x, y, 'k+')
            fig.colorbar(cax, ax=ax, label='Conductivity [mS/m]')
        else:
    #        ax.plot(vertices[:,0], vertices[:,1], 'k.')
            coll = PolyCollection(coordinates, array=sig.flatten('F'), cmap=cmap)
            coll.set_clim(vmin=vmin, vmax=vmax)
            ax.add_collection(coll)
            fig.colorbar(coll, label='Conductivity [mS/m]', ax=ax)

        ax.set_xlabel('X position')
        ax.set_ylabel('Depth [m]')
        ax.set_title(self.surveys[index].name)
        ax.set_ylim([-maxDepth, 0])
        ax.set_xlim([x[0], x[-1]+1])
#        ax.set_aspect('equal')
        def format_coord(i,j):
            col=int(np.floor(i))
            if col < sig.shape[0]:
                row = int(np.where(-depths[col,:] < j)[0].min())-1
                return 'x={0:.4f}, y={1:.4f}, value={2:.4f}'.format(col, row, sig[col, row])
            else:
                return ''
        ax.format_coord = format_coord
        fig.tight_layout()


    def setModels(self, depths, models):
        """Set the models (depth-specific EC) and the depths. Use
        for forward modelling.
        
        Parameters
        ----------
        depths : list of array
            List of array. Each array is npos x ndepths. Each depth
            is a positive number representing the bottom of each layer.
            There is not depth 0.
        models : list of array
            List of array of size npos x nlayers. nlayer should be equals
            exactly to ndepths + 1. All specified in mS/m.
        """
        if len(depths) != len(models):
            raise ValueError('depths and models list should have the same length.')
        for m, d in zip(models, depths):
            if m.shape[1] != d.shape[1] + 1:
                raise ValueError('Number of layer should be equal to number of depths + 1 exactly.')
            if m.shape[0] != d.shape[0]:
                raise ValueError('Number of position in layers and depths array should match.')
        self.depths = depths
        self.models = models

    
    def invertChange(self, **kwargs):
        """Compute depth-specific change in EC based on the difference in ECa
        from the first survey (considered as reference survey). The inversion
        is performed using `invertGN()`.
        """
        self.surveys = []
        for s in self.surveys[1:]:
            s.df[:,self.coils] = s.df[self.coils].values - self.surveys[0].df[self.coils].values
        self.invertGN(**kwargs)
        
    
    def computeChange(self, ref=0):
        """Compute the difference between inverted models compared
        to the ref model.
        
        Parameters
        ----------
        ref : int, optional
            Index of the reference model. By defaut the first model
            (corresponding to the first survey) is used as reference.
        """
        m0 = self.models[0]
        for m in self.models:
            m = m - m0
        
    
    def getRMSE(self):
        """Returns RMSE for all coils (columns) and all surveys (row).
        """
        dfsForward = self.forward()
        def rmse(x, y):
            return np.sqrt(np.sum((x - y)**2)/len(x))
        
        dfrmse = pd.DataFrame(columns=np.r_[self.coils, ['all']])
        for i in range(len(self.surveys)):
            survey = self.surveys[i]
            for coil in self.coils:
                obsECa = survey.df[coil].values
                simECa = dfsForward[i][coil].values
                dfrmse.loc[i, coil] = rmse(obsECa, simECa)
            obsECa = survey.df[self.coils].values.flatten()
            simECa = dfsForward[i][self.coils].values.flatten()
            dfrmse.loc[i, 'all'] = rmse(obsECa, simECa)
        
        return dfrmse
        
        
        
    def showMisfit(self, index=0, coil='all', ax=None):
        """Show Misfit after inversion.
            
        Parameters
        ----------
        index : int, optional
            Index of the survey to plot.
        coil : str, optional
            Which coil to plot. Default is all.
        ax : matplotlib.Axes, optional
            If specified the graph will be plotted on this axis.
        """
        dfsForward = self.forward()
        survey = self.surveys[index]
        cols = survey.coils
        obsECa = survey.df[cols].values
        simECa = dfsForward[index][cols].values
        if ax is None:
            fig, ax = plt.subplots()
        xx = np.arange(survey.df.shape[0])
        ax.plot(xx, obsECa, '.')
        ax.set_prop_cycle(None)
        ax.plot(xx, simECa, '^-')
        ax.legend(cols)
        ax.set_xlabel('Measurements')
        ax.set_ylabel('Conductivity [mS/m]')
        ax.set_title('Dots (observed) vs triangles (modelled)')
        
        
        
    def showOne2one(self, index=0, coil='all', ax=None, vmin=None, vmax=None):
        """Show one to one plot with inversion results.
            
        Parameters
        ----------
        index : int, optional
            Index of the survey to plot.
        coil : str, optional
            Which coil to plot. Default is all.
        ax : matplotlib.Axes, optional
            If specified the graph will be plotted on this axis.
        vmin : float, optional
            Minimum ECa on the graph.
        vmax : float, optional
            Maximum ECa on the graph.
        """
        dfsForward = self.forward()
        survey = self.surveys[index]
        cols = survey.coils
        obsECa = survey.df[cols].values
        simECa = dfsForward[index][cols].values
        #print('number of nan', np.sum(np.isnan(obsECa)), np.sum(np.isnan(simECa)))
        rmse = np.sqrt(np.sum((obsECa.flatten() - simECa.flatten())**2)/len(obsECa.flatten()))
        rmses = np.sqrt(np.sum((obsECa - simECa)**2, axis=0)/obsECa.shape[0])
        if vmin is None:
            vmin = np.nanpercentile(obsECa.flatten(), 5)
        if vmax is None:
            vmax = np.nanpercentile(obsECa.flatten(), 95)
        if ax is None:
            fig, ax = plt.subplots()
        ax.set_title('RMSE: {:.3f}'.format(rmse))
        ax.plot(obsECa, simECa, '.')
        ax.plot([vmin, vmax], [vmin, vmax], 'k-', label='1:1')
        ax.set_xlim([vmin, vmax])
        ax.set_ylim([vmin, vmax])
        ax.set_xlabel('Observed ECa [mS/m]')
        ax.set_ylabel('Simulated ECa [mS/m]')
        ax.legend(['{:s} ({:.2f})'.format(c, r) for c, r in zip(cols, rmses)])
    
    
    def keepBetween(self, vmin=None, vmax=None):
        """Filter out measurements that are not between vmin and vmax.
        
        Parameters
        ----------
        vmin : float, optional
            Minimal ECa value, default is minimum observed.
        vmax : float, optional
            Maximum ECa value, default is maximum observed.
        """
        for s in self.surveys:
            s.keepBetween(vmin=vmin, vmax=vmax)
    
    
    
    def lcurve(self, isurvey=0, irow=0, alphas=None, ax=None):
        """Compute an L-curve given different values of alphas.
        
        Parameters
        ----------
        isurvey : int, optional
            Index of survey to be used, by default the first one.
        irow : int, optional
            Index of measurements to be used inside the survey. First by default.
        alpha : list or array-like, optional
            List or array of values of alphas to build the L-curve.
        ax : matplotlib.Axes, optional
            If specified, the graph will be plotted agains this axis.
        """
        app = self.surveys[isurvey].df[self.coils].values[irow,:]
        if alphas is None:
            alphas = np.logspace(-3,2,20)
        def fmodel(p):
            return fCS(p, self.depths0, self.cspacing, self.cpos)
        L = buildSecondDiff(len(self.conds0))
        def dataMisfit(p, app):
            return fmodel(p) - app
        def modelMisfit(p):
            return np.dot(L, p)
        def objfunc(p, app, alpha):
            return np.sqrt(np.sum(dataMisfit(p, app)**2)/len(app)
                           + alpha*np.sum(modelMisfit(p)**2)/len(p))
        phiData = np.zeros(len(alphas))
        phiModel = np.zeros(len(alphas))
        for i, alpha in enumerate(alphas):
            res = minimize(objfunc, self.conds0, args=(app, alpha))
            phiData[i] = np.sum(dataMisfit(res.x, app)**2)
            phiModel[i] = np.sum(modelMisfit(res.x)**2)
            
        if ax is None:
            fig, ax = plt.subplots()
        ax.set_title('L curve')
        ax.plot(phiModel, phiData, '.-')
        for a, ix, iy in zip(alphas, phiModel, phiData):
            ax.text(ix, iy, '{:.2f}'.format(a))
        ax.set_xlabel('Model Misfit ||L$\sigma$||$^2$')
        ax.set_ylabel('Data Misfit ||$\sigma_a - f(\sigma)$||$^2$')



    def calibrate(self, fnameECa, fnameEC, forwardModel='CS', ax=None):
        """Calibrate ECa with given EC profile.
        
        Parameters
        ----------
        fnameECa : str
            Path of the .csv file with the ECa data collected on the calibration points.
        fnameEC : str
            Path of the .csv file with the EC profile data. One row per location
            corresponding to the rows of fnameECa. The header should be the
            corresponding depths in meters positive downards.
        forwardModel : str, optional
            Forward model to use. Either CS (default), FS or FSandrade.
        ax : matplotlib.Axes
            If specified the graph will be plotted against this axis.
        """
        survey = Survey(fnameECa)
        if survey.freqs[0] is None: # fallback in case the use doesn't specify the frequency in the headers
            try:
                survey.freqs = np.ones(len(survey.freqs))*self.freqs[0]
                print('EMI frequency not specified in headers, will use the one from the main data:' + str(self.freqs[0]) + 'Hz')
            except:
                print('Frequency not found, revert to CS')
                forwardModel = 'CS' # doesn't need frequency
        dfec = pd.read_csv(fnameEC)
        if survey.df.shape[0] != dfec.shape[0]:
            raise ValueError('input ECa and inputEC should have the same number of rows so the measurements can be paired.')
        depths = np.abs(dfec.columns.values.astype(float)) # those are the depths of at mid layer
        depths = depths[:-1] + np.diff(depths) # those are depths of the bottom of the layer
        
        # define the forward model
        if forwardModel == 'CS':
            def fmodel(p):
                return fCS(p, depths, survey.cspacing, survey.cpos, hx=survey.hx[0])
        elif forwardModel == 'FS':
            def fmodel(p):
                return fMaxwellECa(p, depths, survey.cspacing, survey.cpos, f=survey.freqs[0], hx=survey.hx[0])
        elif forwardModel == 'FSandrade':
            def fmodel(p):
                return fMaxwellQ(p, depths, survey.cspacing, survey.cpos, f=survey.freqs[0], hx=survey.hx[0])
    
        # compute the forward response
        simECa = np.zeros((dfec.shape[0], len(survey.coils)))
        for i in range(dfec.shape[0]):
            simECa[i,:] = fmodel(dfec.values[i,:])
        
        # graph
        obsECa = survey.df[survey.coils].values
        if ax is None:
            fig, ax = plt.subplots()
        ax.plot(obsECa, simECa, '.')
        x = np.r_[obsECa.flatten(), simECa.flatten()]
        x = x[~np.isnan(x)]
        vmin, vmax = np.nanmin(x), np.nanmax(x)
        ax.plot([vmin, vmax], [vmin, vmax], 'k-', label='1:1')
        ax.set_xlim([vmin, vmax])
        ax.set_ylim([vmin, vmax])
        ax.set_xlabel('ECa(EM) [mS/m]')
        ax.set_ylabel('ECa(ER) [mS/m]')
        ax.legend(survey.coils)
        
        # plot equation, apply it or not directly
        predECa = np.zeros(obsECa.shape)
        for i, coil in enumerate(survey.coils):
            x, y = obsECa[:,i], simECa[:,i]
            inan = ~np.isnan(x)& ~np.isnan(y)
            slope, intercept, r_value, p_value, std_err = linregress(x[inan], y[inan])
            print(coil, '{:.2f} * x + {:.2f} (R={:.2f})'.format(slope, intercept, r_value))
            predECa[:,i] = obsECa[:,i]*slope + intercept
        ax.set_prop_cycle(None)
        ax.plot(obsECa, predECa, '-')
        
        
    def crossOverPoints(self, index=0, coil=None, ax=None, dump=print):
        """ Build an error model based on the cross-over points.
        
        Parameters
        ----------
        index : int, optional
            Survey index to fit the model on. Default is the first.
        coil : str, optional
            Name of the coil.
        ax : Matplotlib.Axes, optional
            Matplotlib axis on which the plot is plotted against if specified.
        dump : function, optional
            Function to print the output.
        """
        survey = self.surveys[index]
        survey.crossOverPoints(coil=coil, ax=ax, dump=dump)
    
    
    
    def plotCrossOverMap(self, index=0, coil=None, ax=None):
        """Plot the map of the cross-over points for error model.
        
        Parameters
        ----------
        index : int, optional
            Survey index to fit the model on. Default is the first.
        coil : str, optional
            Name of the coil.
        ax : Matplotlib.Axes, optional
            Matplotlib axis on which the plot is plotted against if specified.
        """
        survey = self.surveys[index]
        survey.plotCrossOverMap(coil=coil, ax=ax)
        
    
    def showSlice(self, index=0, islice=0, contour=False, vmin=None, vmax=None,
                  cmap='viridis_r', ax=None, pts=True):
        """Show depth slice.
        
        Parameters
        ----------
        index : int, optional
            Survey index. Default is first.
        islice : int, optional
            Depth index. Default is first depth.
        contour : bool, optional
            If `True` then there will be contouring.
        vmin : float, optional
            Minimum value for colorscale.
        vmax : float, optional
            Maximum value for colorscale.
        cmap : str, optional
            Name of colormap. Default is viridis_r.
        ax : Matplotlib.Axes, optional
            If specified, the graph will be plotted against it.
        pts : boolean, optional
            If `True` (default) the data points will be plotted over the contour.
        """
        z = self.models[index][:,islice]
        x = self.surveys[index].df['x'].values
        y = self.surveys[index].df['y'].values
        if ax is None:
            fig, ax = plt.subplots()
        else:
            fig = ax.figure
        if vmin is None:
            vmin = np.nanmin(z)
        if vmax is None:
            vmax = np.nanmax(z)
        if contour is False or (y == y[0]).all() or (x == x[0]).all(): # can not contour if they are on a line
            if contour is True:
                print('All points on a line, can not contour this.')
            cax = ax.scatter(x, y, c=z, cmap=cmap, vmin=vmin, vmax=vmax)
        else:
            levels = np.linspace(vmin, vmax, 7)
            cax = ax.tricontourf(x, y, z, levels=levels, cmap=cmap, extend='both')
            if pts:
                ax.plot(x, y, 'k+')
        ax.set_xlabel('x [m]')
        ax.set_ylabel('y [m]')
        fig.colorbar(cax, ax=ax, label='Conductivity [mS/m]')
        depths = np.r_[[0], self.depths0, [-np.inf]]
        ax.set_title('{:.2f}m - {:.2f}m'.format(depths[islice], depths[islice+1]))
        
        
#%%
if __name__ == '__main__':
    # cover crop example
    k = Problem()
    k.depths0 = np.array([0.2, 1]) # not starting at 0 !
    k.conds0 = np.ones(len(k.depths0)+1)*20
    k.createSurvey('test/coverCrop.csv', freq=30000)
#    k.convertFromNMEA()
#    k.createSurvey('test/warren170316.csv', freq=30000)
    k.surveys[0].df = k.surveys[0].df[:10]
#    k.show()
#    k.lcurve()
#    k.invertGN(alpha=0.07)
    k.invert(method='CG')
#    k.invert(forwardModel='CS', alpha=0.07, method='L-BFGS-B', options={'maxiter':100}, beta=0, fixedDepths=False) # this doesn't work well
#    k.invertGN() # similar as CG with nit=2
#    k.invertQ()
#    k.showMisfit()
#    k.showResults(vmin=30, vmax=50)
    k.showOne2one()
#    k.showMisfit()
#    k.models[0] = np.ones(k.models[0].shape)*20
#    k.forward(forwardModel='FSandrade')
#    k.calibrate('test/dfeca.csv', 'test/dfec.csv', forwardModel='FS') # TODO
    
#    k.showSlice(contour=False, cmap='jet', vmin=10, vmax=50)
    
    #%% test for inversion with FSandrade
    import time
    cond = np.array([10, 20, 30, 30])
#    app = fMaxwellQ(cond, k.depths0, k.cspacing, k.cpos, hx=k.hx[0], f=k.freqs[0])
#    app = fMaxwellECa(cond, k.depths0, k.cspacing, k.cpos, hx=k.hx[0], f=k.freqs[0])
    app = k.surveys[0].df[k.coils].values[0,:]
    L = buildSecondDiff(len(cond))
    def objfunc(p, app):
        return np.sqrt((np.sum((app - fMaxwellECa(p, k.depths0, k.cspacing, k.cpos, hx=k.hx[0], f=k.freqs[0]))**2)
                              + 0.07*np.sum(np.dot(L, p[:,None])**2))/len(app))
    t0 = time.time()
    res = minimize(objfunc, k.conds0, args=(app,), method='Nelder-Mead', options={'maxiter':10})
    print(res)
    print('{:.3f}s'.format(time.time() - t0))
    
    #%%
    solvers = ['Nelder-Mead', 'Powell', 'CG', 'BFGS',
               'L-BFGS-B', 'TNC', 'COBYLA', 'SLSQP']
    tt = []
    for solver in solvers:
        print(solver)
        t0 = time.time()
        res = minimize(objfunc, k.conds0, args=(app,), method=solver)
        tt.append([time.time() - t0, res.nfev, res.fun])
    
    tt = np.vstack(tt)
    xx = np.arange(len(solvers))
    fig, ax = plt.subplots()
    ax.plot(xx, tt)
    ax.set_xticks(xx)
    ax.set_xticklabels(solvers, rotation=90)
    fig.tight_layout()
    fig.show()
    
#%%   see if we can speed-up minimize by decreasing iteration number
    nits = [0, 1, 2, 3, 10, 20, 50, 100, 1000]
    for nit in nits:
        t0 = time.time()
        res = minimize(objfunc, k.conds0, args=(app,), method='Nelder-Mead',
                       options={'maxiter':nit})
        elapsed = time.time() - t0
        print(res.x)
        rmseEC = np.sqrt(np.sum((res.x - cond)**2)/len(cond))
        rmseECa = np.sqrt(np.sum((fMaxwellECa(res.x, k.depths0, k.cspacing, k.cpos, hx=k.hx[0], f=k.freqs[0]) - app)**2)/len(app))
        print('nit={:d} in {:.3f}s with RMSE={:.2f}'.format(nit, elapsed, rmseECa))
    
    
    # mapping example (potatoes)
#    k = Problem()
#    k.createSurvey('test/regolith.csv')
#    k.convertFromNMEA()
#    k.showMap(contour=True, pts=True)
#    k.show()
#    k.gridData(method='cubic')
#    k.surveys[0].df = k.surveys[0].dfg
#    k.showMap(coil = k.coils[1])
#    

#%% GF direct import
#    k = Problem()
#    k.importGF('emagpy/test/coverCropLo.dat', 'emagpy/test/coverCropHi.dat')
#    k.invertGN()
#    k.showSlice(contour=True)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Apr 24 07:52:50 2019

@author: jkl
"""
#from numba import njit, jit
import os
import numpy as np
#import matplotlib.pyplot as plt
#from multiprocessing import Pool, Manager
from scipy.constants import mu_0
#from hankel import HankelTransform
from scipy import special, integrate
from scipy.optimize import newton#, brent, minimize, curve_fit,


# useful functions for Hankel transform

# load this important variables once (to speed up) -> do not overwrite them !
dirname = os.path.dirname(__file__)
print(dirname)
hankel_w0 = np.loadtxt(os.path.join(dirname, 'j1_140.txt'))
hankel_w1 = np.loadtxt(os.path.join(dirname, 'j0_120.txt'))
hankel5_w0 = np.loadtxt(os.path.join(dirname, 'hankelwts0.txt'))
hankel5_w1 = np.loadtxt(os.path.join(dirname, 'hankelwts1.txt'))
hankel5_lamb = np.loadtxt(os.path.join(dirname, 'hankelpts.txt'))


def func_hankel(typ, K, r):
    ''' Hankel transform based on Guptasarma et al.
    '''
    if typ == 0:
        a = -8.3885
        s = 0.0904
        i = np.arange(0,120).reshape((1,120))
#        w = np.loadtxt('j0_120.txt')
        w = hankel_w0
    elif typ == 1:
        a = -7.9100
        s = 0.0880
        i = np.arange(0,140).reshape((1,140))
#        w = np.loadtxt('j1_140.txt')
        w = hankel_w1
    lamb = 1/r*10**(a+(i-1)*s) # lambda here is not the frequency just your integral variable
#    K0 = np.exp(-lamb0*(z+htx)+getRTE(lamb0)*np.exp(lamb0*(z-htx)))*lamb0    
    hankel = np.sum(K(lamb)*w)/r
    return hankel


def func_hankel2(nu, f, s, N=140, h=0.03):
    """ This one is from hankel python package and allow to
    adjust N and h.
    """
    ht = HankelTransform(nu, N, h) # for HCP h=0.2 works well
    Fk = ht.transform(f, k=s)
    return Fk[0]


def func_hankel3(nu, f, s):
    """ Manual integration using scipy integrate
    """
    def integrand(x):
        return f(x)*special.jv(nu, s*x)*x
    return integrate.quad(integrand, 0, 100)


def func_hankel5(typ, K, r): # prefered one as the step is not linear >< hankel
    ''' we need this type of hankel transform because we can input r (Tx-Rx)
    coil spacing. We can't do that with the python hankel package.
    based on Anderson Fortran code 1979 Geophysic 44(7): 1287-1305
    '''
    if typ == 0:
#        w = np.loadtxt('hankelwts0.txt')
        w = hankel5_w0
    elif typ == 1:
#        w = np.loadtxt('hankelwts1.txt')
        w = hankel5_w1
#    lamb = np.loadtxt('hankelpts.txt') # lambda here is not the frequency just your integral variable
    lamb = hankel5_lamb
    hankel = np.sum(w*K(lamb/r)/r)
    return hankel

""" NOTE
func_hankel and func_hankel5 do not contains the multiplication by lambda in
in their kernel function. This has to be added after.
"""



# useful functions for full Maxwell

def getR0_1(lamb, sigg, f, d):
    """ Compute reflexion coefficient based on Deidda 2018.
    
    Parameters
    ----------
    lamb : numpy.array
        Radial wave number (integration variable of Hankel transform).
    sigg : numpy.array
        Conductivity of the n layers.
    f : float
        Frequency in Hz of the instrument.
    d : numpy.array
        Height of each layers [m].
    
    Returns
    -------
    R0 : complex
        Reflexion coefficient for layer 0 (air layer).
    """
    sigma = np.r_[0, sigg.copy()] # sigma 0 is above the ground
    d = np.r_[d.copy(), 0] # last one won't be used, just for index
    omega = 2*np.pi*f # assume single omega here as lamb is already a vector
    u = np.sqrt(lamb**2+1j*sigma*mu_0*omega)
    N = u/(1j*mu_0*omega)
    Y = np.zeros(N.shape, dtype=np.complex)*np.nan # Y is actually -1 smaller than N but easier like this for index
    Y[-1] = N[-1]
    for i in range(len(N)-2, 0, -1):
        Y[i] = N[i]*(Y[i+1]+N[i]*np.tanh(d[i]*u[i]))/ \
                    (N[i] + Y[i+1]*np.tanh(d[i]*u[i]))
    R0 = (N[0]-Y[1])/(N[0]+Y[1])
    return R0

#@jit(nopython=True) # cannot guessed the type of R (complex)
def getR0_2(lamb, sigg, f, d):
    """ Compute reflexion coefficient based on vonHebel2014, Lavoué2010.
    
    Parameters
    ----------
    lamb : numpy.array
        Radial wave number (integration variable of Hankel transform).
    sigg : numpy.array
        Conductivity of the n layers.
    f : float
        Frequency in Hz of the instrument.
    d : numpy.array
        Height of each layers [m].
    
    Returns
    -------
    R0 : complex
        Reflexion coefficient for layer 0 (air layer).
    """
    sigma = np.concatenate([[0], sigg]) # sigma 0 is above the ground
    h = np.concatenate([[0], d, [0]]) # last one won't be used, because Rn+1 = 0 and
    # first zero won't be used because we never use hn, always hn+1, just
    # for easier indexing
    gamma = np.sqrt(lamb**2 + 1j*2*np.pi*f*mu_0*sigma)
    R = np.zeros((len(sigma)), dtype=np.complex)#*np.nan
#    R[-1] = 0 # no waves from the lower half space
    for i in range(len(sigma)-2, -1, -1):
        num = (gamma[i]-gamma[i+1])/(gamma[i]+gamma[i+1]) \
            + R[i+1]*np.exp(-2*gamma[i+1]*h[i+1])
        den = 1+(gamma[i]-gamma[i+1])/(gamma[i]+gamma[i+1]) \
            * R[i+1]*np.exp(-2*gamma[i+1]*h[i+1])
        R[i] = num/den
    return R[0]

#%timeit getR0_2(1, np.array([60,70,60]), 30000, np.array([0.3, 0.5]))
# 53 us

#@jit
#def foo(sigg, d, lamb):
#    sigma = np.array([0] + sigg)
#    h = np.array([0] + d + [0])
#    gamma = np.sqrt(lamb**2 + 1j*2*np.pi*f*mu_0*sigma)
#    return gamma
#
#%timeit foo([10, 20], [2], 0.3)
##foo.inspect_types()


# test
#%timeit print(getR0_1(1, [60,70,60], 30000, [0.3, 0.5])) # very good exactly the same !
#print(getR0_2(1, np.array([60,70,60]), 30000, np.array([0.3, 0.5])))
# if homogenous, only R0 is != 0, all other are = 0 because there
# is no reflexion if the layers have the same conductivity.

#getRn = np.vectorize(getR0_1)
#getRn = np.vectorize(getR0_2)

def getRn(lamb, sig, f, h, getR0=getR0_2):
    return np.array([getR0(l, sig, f, h) for l in lamb.flatten()])

def getQhomogeneous(cpos, s, sig, f):
    """ Compute quadrature ratio based on analytical solution of McNeil1980
    for homogeneous half-space only.
    """
    omega = 2*np.pi*f
    g = np.sqrt(1j*omega*mu_0*sig) # for gamma
    if cpos == 'hcp':
        Q = 2/(g*s)**2*(9-(9+9*g*s+4*(g*s)**2+(g*s)**3)*np.exp(-g*s))
    elif cpos == 'vcp':
        Q = 2*(1-3/(g*s)**2+(3+3*g*s+(g*s)**2)*np.exp(-g*s)/(g*s)**2)
    return Q
    
    
def getQ(cpos, s, sig, f, h, typ=5):
    """ Compute the quadrature ratio.
    """
    if cpos == 'vcp':
        if typ == 1|5:
            def func(lamb):
                return getRn(lamb, sig, f, h)*lamb
        else:
            def func(lamb):
                return getRn(lamb, sig, f, h)
    elif cpos == 'hcp':
        if typ == 1|5:
            def func(lamb):
                return getRn(lamb, sig, f, h)*lamb**2
        else:
            def func(lamb):
                return getRn(lamb, sig, f, h)*lamb
    dicoHankel = {1:func_hankel,
                2:func_hankel2,
                3:func_hankel3,
                5:func_hankel5}
    dicoBessel = {'hcp':0,
                  'vcp':1}
    dicoCoefs = {'hcp': s**3,
                 'vcp': s**2}
    return 1-dicoCoefs[cpos]*dicoHankel[typ](dicoBessel[cpos], func, s)


def getLIN(sigma, s, f=30000):
    """ Get Low Induction Number. It should be << 1.
    """
    skinDepth = np.sqrt(2/(2*np.pi*f*mu_0*sigma))
    return s/skinDepth


def Q2eca(Q, s, f=30000):
    """ Returns apparent conductivity given the quadrature value (assuming we
    are in the Low Induction Number (LIN) approximation)
    """
    return 4/(2*np.pi*f*mu_0*s**2)*np.imag(Q)

def eca2Q(eca, s, f=30000):
    """ Returns the imaginary part of quadrature given the apparent 
    conductivity value (assuming we are in the Low Induction Number (LIN) 
    approximation)
    """
    return eca/(4/(2*np.pi*f*mu_0*s**2))

# test
#sig = np.array([60,60,60])*1e-3
#f = 30000
#h = [0.3, 0.5]
#s = 0.32
#cpos = 'hcp'
#def func(lamb):
#    return getRn(lamb, sig, f, h)*lamb**2
#app = -4*s/(2*np.pi*30000*mu_0)*np.imag(func_hankel5(0, func, s))
#print(app*1e3)
#print(Q2eca(getQ(cpos, s, sig, f, h), s, f)*1e3) # pk


# test homogeneous
#sigH = 60*1e-3 # S/m
#sig = np.ones(3)*sigH
#f = 30000
#h = [0.3, 0.5]
#s = 0.71
#cpos = 'hcp'
#print(getQ(cpos, s, sig, f, h))
#print(getQhomogeneous(cpos, s, sigH, f)) # not quite...
#print(Q2eca(getQ(cpos, s, sig, f, h), s, f)*1e3)
#print(Q2eca(getQhomogeneous(cpos, s, sigH, f), s, f)*1e3) # not quite...
#



# --------------------- getQ2 faster
#@njit
def getRn2(lamb, sigg, f, d): # compute reflexion coefficients
    sigma = np.array([0] + sigg.tolist()) # sigma 0 is above the ground
    h = np.array([0] + d.tolist() + [0]) # last one won't be used, because Rn+1 = 0 and
    # first zero won't be used because we never use hn, always hn+1, just
    # for easier indexing
    sigmaLen = len(sigma)
    realPart = np.repeat(lamb[:,None], sigmaLen, axis=1)**2
    imagPart = 2*np.pi*f*mu_0*np.repeat(sigma[:,None], len(lamb), axis=1)
    gamma = np.sqrt(realPart.T + 1j*imagPart)
#    gamma = np.sqrt(lamb**2 + 1j*2*np.pi*f*mu_0*sigma)
    ''' optimization idea would be to create the big array gamma before hand
    '''
    R = np.zeros((sigmaLen, len(lamb)), dtype=np.complex)#*np.nan
    R[-1,:] = 0+0j # no waves from the lower half space
    for i in range(sigmaLen-2, -1, -1): # it's recursive so needs for loop
        R[i,:] = (gamma[i,:]-gamma[i+1,:])/(gamma[i,:]+gamma[i+1,:]) \
            + R[i+1,:]*np.exp(-2*gamma[i+1,:]*h[i+1])
        R[i,:] /= 1+(gamma[i,:]-gamma[i+1,:])/(gamma[i,:]+gamma[i+1,:]) \
            * R[i+1,:]*np.exp(-2*gamma[i+1,:]*h[i+1])

    return R[0,:]

# test code
#getR0_2(1, np.array([60,70,60]), 30000, np.array([0.3, 0.5]))
#lamb = np.array([1])
#%timeit getRn2(lamb, np.array([60,60,60]), 30000, np.array([0.3, 0.5]))

#@njit
def getQ2(cpos, s, sig, f, h, typ=None):
    lamb = hankel5_lamb/s # normalized to be used in Hankel
    if cpos == 'vcp':
        w = hankel5_w1 # for Bessel function of order 1
        K = getRn2(lamb, sig, f, h)*lamb # kernel with reflexion coef
        hankel = np.sum(w*K/s) # hankel transform
        return 1-s**2*hankel
    elif cpos == 'hcp':
        w = hankel5_w0 # for Bessel function of order 0
        K = getRn2(lamb, sig, f, h)*lamb**2 # kernel with reflexion coef
        hankel = np.sum(w*K/s) # hankel transform
        return 1-s**3*hankel
    elif cpos == 'prp': # see Anderson for correct formula
        w = hankel5_w1 # for Bessel function of order 0
        K = getRn2(lamb, sig, f, h)*lamb**2 # kernel with reflexion coef
        hankel = np.sum(w*K/s) # hankel transform
        return 0-s**3*hankel
  
def getQs(cond, depths, s, cpos, f, hx=0):
    app = np.zeros(len(cpos), dtype=complex)
    h = np.r_[depths[0], np.diff(depths)]
    if hx > 0:
        h = np.r_[hx,h]
        cond = np.r_[0, cond]
    cond = cond*1e-3
    for i in range(len(cpos)):
        app[i] = getQ2(cpos[i], s[i], cond, f=f, h=h)
    return app


#sigH = 60*1e-3 # S/m
#sig = np.ones(3)*sigH
#f = 30000
#h = np.array([0.3, 0.5])
#s = 0.71
#cpos = 'hcp'
#lamb = hankel5_lamb/s
#%timeit Ka = getRn(lamb, sig, f, h) # 55 ms
#%timeit Kb = getRn2(lamb, sig, f, h) # 662 us
#%timeit getQ(cpos, s, sig, f, h) # 47.5 ms
#getQ2(cpos, s, sig, f, h) # 672 us !
#%timeit getQhomogeneous('hcp',0.71, 60, 30000)
#print(getQ2('hcp', 0.71, np.array([60, 60]), 30000, np.array([1])))
#print(getQ('hcp', 0.71, np.array([60, 60]), 30000, np.array([1])))


# definition of forward model used
def fMaxwellECa(cond, depths, s, cpos, hx=None, f=30000):
    """ Compute quadrature ratio Q using Maxwell equation but then use the LIN
    to convert it to apparent conductivity.
    """
    if len(cond)-1 != len(depths):
        raise ValueError('len(cond)-1 should be equal to len(depths).')
    if len(s) != len(cpos):
        raise ValueError('len(s) should be equal to len(cpos).')
    h = np.r_[depths[0], np.diff(depths)]
    if hx is not None:
        h = np.r_[hx, h]
        cond = np.r_[0, cond]
    response = np.zeros(len(s))*np.nan
    cond = cond*1e-3 # convert to S/m
    for i in range(len(s)):
        Q = getQ2(cpos[i], s[i], cond, f, h)
        response[i] = Q2eca(Q, s[i], f)*1e3
    return response


def fMaxwellQ(cond, depths, s, cpos, hx=None, f=30000, maxiter=50):
    """ Compute the quadrature ratio using Maxwell equation and then try to
    find a value of conductivity that match the observed Q
    """
    if len(cond)-1 != len(depths):
        raise ValueError('len(cond)-1 should be equal to len(depths).')
    if len(s) != len(cpos):
        raise ValueError('len(s) should be equal to len(cpos).')
#    eca0 = fCS(cond, depths, s, cpos, hx=hx) # to kick start the solver
    h = np.r_[depths[0], np.diff(depths)]
    if hx is not None:
        h = np.r_[hx, h]
        cond = np.r_[0, cond]
    response = np.zeros(len(s))*np.nan
    cond = cond*1e-3 # convert to S/m
    for i in range(len(s)):
        Qobs = np.imag(getQ2(cpos[i], s[i], cond, f, h))
        if cpos[i] == 'prp': # we don't have analytical for PRP
            def objfunc(sig):
                sigg = np.ones(len(cond))*sig
                Qmod = np.imag(getQ2(cpos[i], s[i], sigg, f, h))
                return np.abs(Qmod - Qobs)
        else:
            def objfunc(sig): # analytical is much faster
                Qmod = np.imag(getQhomogeneous(cpos[i], s[i], sig, f)) # faster even if more rounding errors
                return np.abs(Qmod - Qobs)
        sig0 = Q2eca(getQ2(cpos[i], s[i], cond, f, h), s[i], f) # still in S/m
#        sig0 = eca0[i]
#        res = minimize(objfunc, x0=sig0, method='BFGS')
#        zero = res.x[0]*1e3
#        zero = brent(objfunc, brack=(sig0*0.8, sig0, sig0*1.2))*1e3
        zero = newton(objfunc, sig0, maxiter=maxiter)*1e3 # back to mS/m # faster !
        response[i] = zero
    return response


# test
#cpos = 'hcp'
#s = 0.32
#cond = np.array([60,60,60])*1e-3 # [S/m]
#f = 30000
#h = [0.4, 0.7]
#print(getQ(cpos, s, cond, f, h))
#print(getQhomogeneous(cpos, s, cond, f))
#Qobs = np.imag(getQ(cpos, s, cond, f, h))
#def objfunc(sig):
#    sigg = np.ones(len(cond))*sig
#    Qmod = np.imag(getQ(cpos, s, sigg, f, h))
#    return np.abs(Qmod - Qobs)
#sig0 = Q2eca(getQ(cpos, s, cond, f, h), s, f) # still in S/m
#print(sig0)
#zero = newton(objfunc, sig0)*1e3 # back to mS/m
#print(zero)



def Q2eca2(Qobs, depths, cpos, s, f=30000, hx=None, maxiter=50):
    Qobs = np.imag(Qobs)
    h = np.r_[depths[0], np.diff(depths)]
    if hx is not None:
        h = np.r_[hx, h]
    if cpos == 'prp': # we don't have analytical for PRP
        def objfunc(sig):
            sigg = np.ones(len(depths)+1)*sig
            Qmod = np.imag(getQ2(cpos, s, sigg, f, h))
            return np.abs(Qmod - Qobs)
    else:
        def objfunc(sig): # analytical is much faster
            Qmod = np.imag(getQhomogeneous(cpos, s, sig, f)) # faster even if more rounding errors
            return np.abs(Qmod - Qobs)
    sig0 = Q2eca(0+1j*Qobs, s) # still in S/m
#    sig0 = eca0[i]
#    res = minimize(objfunc, x0=sig0, method='BFGS')
#    zero = res.x[0]*1e3
#    zero = brent(objfunc, brack=(sig0*0.8, sig0, sig0*1.2))*1e3
    zero = newton(objfunc, sig0, maxiter=maxiter)
    return zero

# test
#Qobs = getQ2('hcp',0.71, np.array([20,20])*1e-3, 30000, np.array([0.5]))
#print(Q2eca(Qobs, 0.71)*1e3)
#print(Q2eca2(Qobs, np.array([0.5]), 'hcp', 0.71, f=30000 )*1e3)



def emSens(depths,s,coilPosition, hx=0, rescaled=False):
    """return mcNeil senstivity values
    
    depths (positive and increasing with depth)
    s = distance between coils
    coilPosition is "hcp" for horizontal coplanar (HCP) and "vcp" for VCP
    x = range on which compute the sensitivity (array_like)
     """
    depths = depths + np.abs(hx)
    z=np.array(depths)/s
    # from mcNeil 1980 in Callegary2007
    if coilPosition == 'hcp':
        cs = 1/np.sqrt(4*z**2+1)
    if coilPosition == 'vcp':
        cs = np.sqrt(4*z**2+1)-2*z
    if coilPosition == 'prp':
        cs = 1-2*z/np.sqrt(4*z**2 + 1)
    
    if (hx != 0) and rescaled is True: # rescale the cumulative sensitivity so that it reaches 1
        cs = cs/cs[0]

    return cs # all but the first one as it is just use to scale everything

# test code
#depths = np.arange(0, 2, 0.1)
#cs1 = emSens(depths, 0.32, 'hcp', hx=0)
#cs2 = emSens(depths, 0.32, 'hcp', hx=1)
#
#fig, ax = plt.subplots()
#ax.plot(cs1, -depths, '-')
#ax.plot(cs2, -depths, '--')
#ax.plot(cs1, -depths+1, ':')
#fig.show()


def forward1d_full(cond, depths, s, cpos, hx=None, rescaled=False):
    if hx is None:
        hx = np.zeros(len(s))
    if len(cond.shape)>1:
#        print('WARNING : need 1 dimension matrix for forward model')
        cond = cond.flatten()
    if depths[0] != 0:
        print('ERROR: first depth should be zero, will add it for you')
        depths = np.r_[0, depths]
    if len(depths) != len(cond):
        raise ValueError('forward1d_full: unmatching depths and conds')
        return
    if np.sum(depths < 0) > 0:
        print('ERROR: depth contains negative depth. Please input only positive depth')
        return
    out = np.zeros(len(s))*np.nan
    for i in range(0,len(out)):
        cs = emSens(depths,s[i], cpos[i], hx[i], rescaled=rescaled)
#        print('CS = ', np.sum(np.diff(cs)))
        appcond = np.sum(cond[:-1]*(cs[:-1]-cs[1:]))
        appcond = appcond + cond[-1]*(cs[-1])
        out[i] = appcond
    return out

# test code
#app1 = forward1d_full(np.array([10,30]),[1],[0.32],['hcp'], hx=[1])
#app2 = forward1d_full(np.array([0,10,30]),[1,2],[0.32],['hcp'], hx=[0])
#print(app1, app2)
    

def fCS(cond, depths, s, cpos, hx=None, rescaled=False):
    """ Use the cumulative sensitivity function of McNeil 1980.
    """
    if len(cond)-1 != len(depths):
        raise ValueError('len(cond)-1 should be equal to len(depths).')
    if len(s) != len(cpos):
        raise ValueError('len(s) should be equal to len(cpos).')
    if depths[0] != 0:
#        print('add 0 for you')
        depths = np.r_[0, depths] # otherwise forward1d_full complains
    if hx is not None:
        hx = np.ones(len(cpos))*hx
    response = forward1d_full(cond, depths, s, cpos, hx=hx, rescaled=rescaled)
    return response


# based on Andrade2018 (general case)
def emSensAndrade(depths,s,coilPosition, hx=0):
    """return RESCALED mcNeil senstivity values (Andrade 2018)
    
    depths (positive and increasing with depth)
    s = distance between coils
    coilPosition is "hcp" for horizontal coplanar (HCP) and "vcp" for VCP
    x = range on which compute the sensitivity (array_like)
     """
    depths = depths
    z=np.array(depths)/s
    if coilPosition == 'hcp':
        cs = ((4*(hx/s)**2+1)**0.5)/(4*(z+hx/s)**2+1)**0.5
    if coilPosition == 'vcp':
        cs = ((4*(z+hx/s)**2+1)**0.5 - 2*(z + hx/s))/((4*(hx/s)**2+1)**0.5 - 2*hx/s)
    return cs

def fCSandrade(cond,depths,s,cpos, hx=None):
    if hx is None:
        hx = 0
    hx = np.ones(len(s))*hx
    if len(cond.shape)>1:
#        print('WARNING : need 1 dimension matrix for forward model')
        cond = cond.flatten()
    if depths[0] != 0:
#        print('ERROR: first depth should be zero, will add it for you')
        depths = np.r_[0, depths]
    if len(depths) != len(cond):
        print('ERROR: unmatching depths and conds')
        return
    if np.sum(depths < 0) > 0:
        print('ERROR: depth contains negative depth. Please input only positive depth')
        return
    out = np.zeros(len(s))*np.nan
    for i in range(0,len(out)):
        cs = emSensAndrade(depths,s[i], cpos[i], hx[i])
        appcond = np.sum(cond[:-1]*(cs[:-1]-cs[1:]))
        appcond = appcond + cond[-1]*(cs[-1])
        out[i] = appcond
    return out


def buildSecondDiff(ndiag):
    x=np.ones(ndiag)
    a=np.diag(x[:-1]*-1,k=-1)+np.diag(x*2,k=0)+np.diag(x[:-1]*-1,k=1)
    a[0,0]=1
    a[-1,-1]=1
    return a


#%% test bench
#sigma = np.array([30, 30, 30, 30]) # layers conductivity [mS/m]
#depths = np.array([0.3, 0.7, 2]) # thickness of each layer (last is infinite)
#f = 30000 # Hz frequency of the coil
#cpos = np.array(['hcp','hcp','hcp','vcp','vcp','vcp']) # coil orientation
##cpos = np.array(['hcp','hcp','hcp','prp','prp','prp']) # coil orientation
#cspacing = np.array([0.32, 0.71, 1.18, 0.32, 0.71, 1.18])
#cspacing = np.array([1.48, 2.82, 4.49, 1.48, 2.82, 4.49]) # explorer
#
#print('fCS:', fCS(sigma, depths, cspacing, cpos)) # 333 us
#print('fMaxwellECa:', fMaxwellECa(sigma, depths, cspacing, cpos)) # 5.7 ms
#print('fMaxwellQ:', fMaxwellQ(sigma, depths, cspacing, cpos)) # 12 ms
#print('fCSandrade:', fCSandrade(sigma, depths, cspacing, cpos, hx=0))
#%timeit getQs(sigma, depths, cspacing, cpos, f) # 5.75 ms (same as fMaxwellECa)

#print('fCS:', fCS(sigma, depths, cspacing, cpos, hx=1))
#print('fCS:', fCS(sigma, depths, cspacing, cpos, hx=1, rescaled=True))
#print('fCSandrade:', fCSandrade(sigma, depths, cspacing, cpos, hx=1))


#%% usefull functions for the direct inversion
def buildJacobian(depths, s, cpos):
    '''Build Jacobian matrix for Cumulative Sensitivity based inversion.
    
    Parameters
    ----------
    cond : array
        Conductivities of the model layer.
    depths : array
        Depths of the lower layer bound.
    s : array
        Array of coil separation [m].
    cpos : array of str
        Array of coil orientation.
    
    Returns
    -------
    A matrix with each row per coil spacing/orientation and each column per
    model parameters (per layer).
    '''
    depths = np.r_[0, depths]
    jacob = np.zeros((len(s),len(depths)))*np.nan
    for i in range(0,len(s)):      
        cs = emSens(depths, s[i], cpos[i])         
        jacob[i,:-1] = cs[:-1]-cs[1:]
        jacob[i,-1] = cs[-1]
    return jacob

# test
#J = buildJacobian([0.5, 0.7],[0.32, 0.72, 1.14],['vcp','vcp','vcp'])



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Jun  7 21:59:16 2019

@author: jkl
"""

import numpy as np
import matplotlib.pyplot as plt
from scipy.optimize import minimize
import time
plt.ioff()
from tools.invertEMFullFunc import fMaxwellQ, fMaxwellECa, fCS
plt.close('all')


def addnoise(x, level=0.02):
    return x + np.random.randn(1) * x * level
addnoise = np.vectorize(addnoise)


def buildSecondDiff(ndiag):
    x=np.ones(ndiag)
    a=np.diag(x[:-1]*-1,k=-1)+np.diag(x*2,k=0)+np.diag(x[:-1]*-1,k=1)
    a[0,0]=1
    a[-1,-1]=1
    return a


#%%
conds = np.array([20, 30, 20, 50])
depths = np.array([0.5, 1, 1.5])
ec0 = 25 # mS/m
conds0 = np.ones(len(conds))*ec0

fmodel = fCS # 0.011 s
#fmodel = fMaxwellECa # 0.099 s
#fmodel = fMaxwellQ # 0.195 s

cpos = ['vcp','vcp','vcp','hcp','hcp','hcp']
cspacing = [0.32, 0.71, 1.18, 0.32, 0.71, 1.18]

# create ECa
app = fCS(conds, depths, cspacing, cpos)
#app = addnoise(app)

L = buildSecondDiff(len(conds))
def objfunc(p, app):
    return np.sqrt((np.sum((app - fmodel(p, depths, cspacing, cpos))**2)
                          + 0.07*np.sum(np.dot(L, p[:,None])**2))/len(app))
t0 = time.time()
res = minimize(objfunc, conds0, args=(app,), method='Nelder-Mead')
print(res)
print('{:.3f}s'.format(time.time() - t0))

    
#%%   see if we can speed-up minimize by decreasing iteration number
nits = [0, 1, 2, 3, 10, 20, 50, 100, 1000]
rmseECs = []
rmseECas = []
elapseds = []
for nit in nits:
    t0 = time.time()
    res = minimize(objfunc, conds0, args=(app,), method='L-BFGS-B',
                   options={'maxiter':nit})
    elapsed = time.time() - t0
    print(res.x)
    rmseEC = np.sqrt(np.sum((res.x - conds)**2)/len(conds))
    rmseECa = np.sqrt(np.sum((fmodel(res.x, depths, cspacing, cpos) - app)**2)/len(app))
    rmseECs.append(rmseEC)
    rmseECas.append(rmseECa)
    elapseds.append(elapsed)
    print('nit={:d} in {:.3f}s with RMSE={:.2f}'.format(nit, elapsed, rmseECa))
    

fig, ax = plt.subplots()
ax.semilogx(nits, rmseECs, '.-', label='RMSE (EC)')
ax.semilogx(nits, rmseECas, '.-', label='RMSE (ECa)')
ax.legend()
ax2 = ax.twinx()
ax2.semilogx(nits, elapseds, 'k.--', label='time (sec)')
ax.set_xlabel('Max Number of Iteration')
ax.set_ylabel('RMSE')
ax2.set_ylabel('Elasped [s]')
fig.tight_layout()
fig.show()

''' 10 iterations is a good trade-off
'''

    #%%
    solvers = ['Nelder-Mead', 'Powell', 'CG', 'BFGS',
               'L-BFGS-B', 'TNC', 'COBYLA', 'SLSQP']
    tt = []
    for solver in solvers:
        print(solver)
        t0 = time.time()
        res = minimize(objfunc, k.conds0, args=(app,), method=solver)
        tt.append([time.time() - t0, res.nfev, res.fun])
    
    tt = np.vstack(tt)
    xx = np.arange(len(solvers))
    fig, ax = plt.subplots()
    ax.plot(xx, tt)
    ax.set_xticks(xx)
    ax.set_xticklabels(solvers, rotation=90)
    fig.tight_layout()
    fig.show()
    


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Apr 16 20:27:12 2019

@author: jkl
"""
import os
from datetime import time, datetime
#NB: using python datetime due to compatbility issues between pandas and matplotlib for plotting
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
from scipy.stats import linregress, binned_statistic
from scipy.spatial.distance import cdist, pdist
from scipy.interpolate import griddata, NearestNDInterpolator
from scipy.spatial import Delaunay
from scipy.spatial import ConvexHull
from emagpy.invertHelper import emSens, Q2eca
#if __name__ == '__main__':
#    import isinpolygon as iip
#else:
#    import emagpy.isinpolygon as iip


def clipConvexHull(xdata,ydata,x,y,z):
    """ set to nan data outside the convex hull of xdata, yxdata
    
    Parameters:
        xdata, ydata (arrays) : x and y position of the data collected
        x, y (arrays) : x and y position of the computed interpolation points
        z (arrays) : value of the interpolation
    
    Return:
        znew (array) : a copy of z with nan when outside convexHull
    """
    knownPoints = np.array([xdata, ydata]).T
    newPoints = np.array([x,y]).T
    _, idx = np.unique(xdata+ydata, return_index=1)
    q = ConvexHull(knownPoints[idx,:])
    hull = Delaunay(q.points[q.vertices, :])
#    qq = q.points[q.vertices,:]
        
    znew = np.copy(z)
    for i in range(0, len(z)):
        if hull.find_simplex(newPoints[i,:])<0:
            znew[i] = np.nan
    return znew


def idw(xnew, ynew, xknown, yknown, zknown, n=1):
    znew = np.zeros(len(xnew))
    for i,(x,y) in enumerate(zip(xnew, ynew)):
#        dist = pdist(x, y, xknown, yknown)
        dist = np.sqrt((x-xknown)**2+(y-yknown)**2)
        # if we consider all the points (might be not good)
        w = (1/dist)**n # exponent to be chosen
        znew[i] = np.sum(zknown*w)/np.sum(w)
    return znew


def convertFromCoord(df, targetProjection='EPSG:27700'):
    """Convert coordinates string (NMEA or DMS) to selected CRS projection.
    
    Parameters
    ----------
    df : pandas.DataFrame
        Dataframe with a 'Latitude' and 'Longitude' columns that contains NMEA
        string.
    targetProjection : str, optional
        Target CRS, in EPSG number: e.g. `targetProjection='EPSG:27700'`
        for the British Grid.
    """
    import pyproj
    
    def NMEA(arg):
        """ Convert NMEA string to WGS84 (GPS) decimal degree.
        """
        letter = arg[-1]
        if (letter == 'W') | (letter == 'S'):
            sign = -1
        else:
            sign = 1
        arg = arg[:-1]
        x = arg.index('.')
        a = float(arg[:x-2]) # degree
        b = float(arg[x-2:]) # minutes
        return (a + b/60)*sign
    
    def DMS(arg):
        """Convert convert degrees, minutes, seconds to decimal degrees
        """
        letter = arg[-1]
        if (letter == 'W') | (letter == 'S'):
            sign = -1
        else:
            sign = 1
        #extract degrees from string
        deg_idx = arg.find('°')
        deg = int(arg[:deg_idx])
        #extract minutes from string
        min_idx = arg.find("'")
        mins = int(arg[(deg_idx+1):min_idx])
        #extract seconds from string
        sec_idx = arg.find('"')
        secs = float(arg[(min_idx+1):sec_idx])
        DD = deg + (mins/60) + (secs/3600) # decimal degree calculation
        return sign*DD # return with sign
    
    check = df['Latitude'][0]
    if check.find('°') !=-1 and check.find("'") != -1:
        print("Coordinates appear to be given as Degrees, minutes, seconds ... adjusting conversion scheme")
        gps2deg = np.vectorize(DMS)
    else:
        gps2deg = np.vectorize(NMEA)
    
    
    df['lat'] = gps2deg(df['Latitude'].values)
    df['lon'] = gps2deg(df['Longitude'].values)
    
    wgs84 = pyproj.Proj("+init=EPSG:4326") # LatLon with WGS84 datum used by GPS units and Google Earth
    osgb36 = pyproj.Proj("+init=" + targetProjection) # UK Ordnance Survey, 1936 datum
    
    df['x'], df['y'] = pyproj.transform(wgs84, osgb36, 
                          df['lon'].values, df['lat'].values)

    return df
    
    
class Survey(object):
    """ Create a Survey object containing the raw EMI data.
    
    
    """
    def __init__(self, fname=None, freq=None, hx=None):
        self.df = None # main dataframe
        self.drift_df = None # drift station dataframe
        self.fil_df = None # filtered dataframe 
        self.freqs = [] # frequency of each coil [Hz]
        self.errorModel = None # a function that returns an error
        self.sensor = None # sensor name (+ company)
        self.coils = [] # columns with the coils names and configuration
        self.cpos = [] # orientation of the coils
        self.cspacing = [] # spacing between Tx and Rx [m]
        self.coilsInph = [] # name of the coils with inphase value in [ppt]
        self.hx = [] # height of the instrument above the ground [m]
        self.name = ''
        self.iselect = []
        if fname is not None:
            self.readFile(fname)
            if freq is not None:
                self.freqs = np.ones(len(self.coils))*freq
            if hx is not None:
                self.hx = np.ones(len(self.coils))*hx
        
    
    def readFile(self, fname, sensor=None):
        """Read a .csv file.
        
        Parameters
        ----------
        fname : str
            Filename.
        sensor : str, optional
            Type of sensor.
        """
        name = os.path.basename(fname)[:-4]
        delimiter=','
        if fname.find('.DAT')!=-1:
            delimiter = '\t'
        df = pd.read_csv(fname, delimiter=delimiter)
        self.readDF(df, name, sensor)
        
        
    def readDF(self, df, name=None, sensor=None):
        """Parse dataframe.
        """
        self.name = 'Survey 1' if name is None else name
        for c in df.columns:
            orientation = c[:3]
            if ((orientation == 'VCP') | (orientation == 'VMD') | (orientation == 'PRP') |
                    (orientation == 'HCP') | (orientation == 'HMD')):
                # replace all orientation in HCP/VCP/PRP mode
                if orientation == 'HMD':
                    df = df.rename(columns={c:c.replace('HMD','VCP')})
                if orientation == 'VMD':
                    df = df.rename(columns={c:c.replace('VMD','HCP')})
                if c[-5:] == '_inph':
                    self.coilsInph.append(c)
                else:
                    self.coils.append(c)
        if 'x' not in df.columns:
            df['x'] = np.arange(df.shape[0])
        if 'y' not in df.columns:
            df['y'] = 0 # maybe not needed
        coilInfo = [self.getCoilInfo(c) for c in self.coils]
        self.freqs = [a['freq'] for a in coilInfo]
        self.hx = [a['height'] for a in coilInfo]
        self.cspacing = [a['coilSeparation'] for a in coilInfo]
        self.cpos = [a['orientation'] for a in coilInfo]
        self.df = df
        self.sensor = sensor

        
    def getCoilInfo(self, arg):
        orientation = arg[:3].lower()
        b = arg[3:].split('f')
        coilSeparation = float(b[0])
        if len(b) > 1:
            c = b[1].split('h')
            freq = float(c[0])
            if len(c) > 1:
                height = float(c[1])
            else:
                height = 0
        else:
            freq = 30000 # Hz default is not specified !!
            height = 0
        return {'orientation': orientation,
                'coilSeparation': coilSeparation,
                'freq': freq,
                'height': height}
        
    
    def keepBetween(self, vmin=None, vmax=None):
        """Filter out measurements that are not between vmin and vmax.
        
        Parameters
        ----------
        vmin : float, optional
            Minimal ECa value, default is minimum observed.
        vmax : float, optional
            Maximum ECa value, default is maximum observed.
        """
        if vmin is not None:
            ie1 = (self.df[self.coils].values > vmin).all(1)
        else:
            ie1 = np.ones(self.df.shape[0], dtype=bool)
        if vmax is not None:
            ie2 = (self.df[self.coils].values < vmax).all(1)
        else:
            ie2 = np.ones(self.df.shape[0], dtype=bool)
        ie = ie1 & ie2
        print('Deleted {:d}/{:d} measurements'.format(np.sum(~ie), len(ie)))
        self.df = self.df[ie]
        
        
    def rollingMean(self, window=3):
        """Perform a rolling mean on the data.
        
        Parameters
        ----------
        window : int, optional
            Size of the windows for rolling mean.
        """
        cols = ['x','y'] + self.coils + self.coilsInph
        self.df[cols] = self.df[cols].rolling(window).mean()
        i2discard = self.df[self.coils].isna().any(1)
        self.df = self.df[~i2discard]
        print('dataset shrunk of {:d} measurements'.format(np.sum(i2discard)))        
    
    
    def filterDiff(self, coil=None, thresh=5):
        """Filter out consecutive measurements when the difference between them
        is larger than val.
        
        Parameters
        ----------
        thresh : float, optional
            Value of absolute consecutive difference above which the second 
            data point will be discarded.
        coil : str, optional
            Coil on which to apply the processing.
        """
        if coil is None:
            coil = self.coils[0]
        val = self.df[coil].values
        i2keep = np.r_[0, np.abs(np.diff(val))] < thresh
        print('filterDiff: {:d}/{:d} measurements removed.'.format(np.sum(~i2keep), len(i2keep)))
        self.df = self.df[i2keep]
    
    
    def show(self, coil='all', attr='ECa', ax=None, vmin=None, 
             vmax=None):
        """ Show the data.
        """
        if coil == 'all':
            cols = self.coils
        else:
            cols = coil
        
        if ax is None:
            fig, ax = plt.subplots()
     
        # interactive point selection
        self.iselect = np.zeros(self.df.shape[0], dtype=bool)
        xpos = np.arange(self.df.shape[0])
        
        def setSelect(ie, boolVal):
            ipoints[ie] = boolVal
            self.iselect = ipoints
    
        def onpick(event):
            xid = xpos[event.ind[0]]
            isame = xpos == xid
            if (ipoints[isame] == True).all():
                setSelect(isame, False)
            else:
                setSelect(isame, True)
            if len(y.shape) == 1:
                killed.set_xdata(x[ipoints])
                killed.set_ydata(y[ipoints])
            else:
                xtmp = []
                ytmp = []
                for i in range(y.shape[1]):
                    xtmp.append(x[ipoints])
                    ytmp.append(y[ipoints, i])
                killed.set_xdata(xtmp)
                killed.set_ydata(ytmp)
            killed.figure.canvas.draw()
    
        ax.set_title(coil)
        caxs = ax.plot(self.df[cols].values, '.-', picker=5)
        ax.legend(cols)
        ax.set_ylim([vmin, vmax])
        ax.set_xlabel('Measurements')
        if coil[-5:] == '_inph':
            ax.set_ylabel('Inphase [ppt]')
        else:
            ax.set_ylabel('Apparent Conductivity [mS/m]')
        for cax in caxs:
            cax.figure.canvas.mpl_connect('pick_event', onpick)        
        killed, = ax.plot([],[],'rx')
        x = xpos
        y = self.df[cols].values
        ipoints = np.zeros(len(x), dtype=bool)


    def dropSelected(self):
        self.df = self.df[~self.iselect]
        print('{:d} points removed'.format(np.sum(self.iselect)))        


    def tcorrECa(self, tdepths, tprofile):
        """Temperature correction using XXXX formula.
        
        Parameters
        ----------
        tdepths : array-like
            Depths in meters of the temperature sensors (negative downards).
        tprofile : array-like
            Temperature values corresponding in degree Celsius.
        """
        # TODO if ECa -> compute an 'apparent' temperature
        # TODO what about per survey ?
        pass


    def convertFromNMEA(self, targetProjection='EPSG:27700'): # British Grid 1936
        """Convert coordinates string (NMEA or DMS) to selected CRS projection.
    
        Parameters
        ----------
        targetProjection : str, optional
            Target CRS, in EPSG number: e.g. `targetProjection='EPSG:27700'`
            for the British Grid.
        """
        self.df = convertFromCoord(self.df, targetProjection)


    
    def showMap(self, coil=None, contour=False, ax=None, vmin=None, vmax=None,
                pts=False, cmap='viridis_r', xlab='x', ylab='y', nlevel=7):
        """ Display a map of the measurements.
        
        Parameters
        ----------
        coil : str, optional
            Name of the coil to plot. By default, the first coil is plotted.
        contour : bool, optional
            If `True` filled contour will be plotted using `tricontourf`.
        ax : Matplotlib.Axes, optional
            If specified, the graph will be plotted against the axis.
        vmin : float, optional
            Minimum of the colorscale.
        vmax : float, optional
            Maximum of the colorscale.
        pts : bool, optional
            If `True` the measurements location will be plotted on the graph.
        xlab : str, optional
            X label.
        ylab : str, optional
            Y label.
        nlevel : int, optional
            Number of levels for the contourmap. Default is 7.
        """
        if coil is None:
            coil = self.coils[0]
#        if coil == 'all': # trick for ui
#            coil = self.coils[-1]
        x = self.df['x'].values
        y = self.df['y'].values
        val = self.df[coil].values
        if ax is None:
            fig, ax = plt.subplots()
        else:
            fig = ax.figure
        if vmin is None:
            vmin = np.nanpercentile(val, 5)
        if vmax is None:
            vmax = np.nanpercentile(val, 95)
        ax.set_title(coil)
        if contour is True:
            levels = np.linspace(vmin, vmax, nlevel)
#            nx = 100
#            ny = 100
#            X, Y = np.meshgrid(np.linspace(np.min(x), np.max(x), nx),
#                           np.linspace(np.min(y), np.max(y), ny))
#            inside = np.ones(nx*ny)
#            inside2 = clipConvexHull(x, y, X.flatten(), Y.flatten(), inside)
#            ie = ~np.isnan(inside2)
#            Z = idw(X.flatten(), Y.flatten(), x, y, val)
##            z = griddata(np.c_[x, y], values, (X, Y), method=method)
#            Z[~ie] = np.nan
#            Z = Z.reshape(X.shape)
#            cax = ax.contourf(X, Y, Z, levels=levels)
#            
            cax = ax.tricontourf(x, y, val, levels=levels, extend='both', cmap=cmap)
            if pts is True:
                ax.plot(x, y, 'k+')
        else:
            cax = ax.scatter(x, y, s=15, c=val, vmin=vmin, vmax=vmax, cmap=cmap)
        ax.set_xlabel(xlab)
        ax.set_ylabel(ylab)
        ax.set_xlim([np.nanmin(x), np.nanmax(x)])
        ax.set_ylim([np.nanmin(y), np.nanmax(y)])
        if coil[-5:] == '_inph':
            fig.colorbar(cax, ax=ax, label='Inphase [ppt]')
        else:
            fig.colorbar(cax, ax=ax, label='Apparent Conductivity [mS/m]')
        

    def saveMap(self, fname, coil=None, nx=100, ny=100, method='linear',
                xmin=None, xmax=None, ymin=None, ymax=None, color=False,
                cmap='viridis_r', vmin=None, vmax=None, nlevel=7):
        """Save a georeferenced raster TIFF file.
        
        Parameters
        ----------
        fname : str
            Path of where to save the .tiff file.
        coil : str, optional
            Name of the coil to plot. By default, the first coil is plotted.
        nx : int, optional
            Number of points in x direction.
        ny : int, optional
            Number of points in y direction.
        method : str, optional
            Interpolation method (nearest, cubic or linear see
            `scipy.interpolate.griddata`) or IDW (default).
        xmin : float, optional
            Mininum X value.
        xmax : float, optional
            Maximum X value.
        ymin : float, optional
            Minimium Y value.
        ymax : float, optional
            Maximum Y value
        color : bool, optional
            If True a colormap will be applied.
        cmap : str, optional
            If `color == True`, name of the colormap. Default is viridis.
        vmin : float, optional
            Minimum value for colomap.
        vmax : float, optional
            Maximum value for colormap.
        nlevel : int, optional
            Number of level in the colormap. Default 7.
        """
        import rasterio
        from rasterio.transform import from_origin
        
        if coil is None:
            coil = self.coils[0]
        xknown = self.df['x'].values
        yknown = self.df['y'].values
        if xmin is None:
            xmin = np.min(xknown)
        if xmax is None:
            xmax = np.max(xknown)
        if ymin is None:
            ymin = np.min(yknown)
        if ymax is None:
            ymax = np.max(yknown)
        X, Y = np.meshgrid(np.linspace(xmin, xmax, nx),
                           np.linspace(ymin, ymax, ny))
        x, y = X.flatten(), Y.flatten()
        values = self.df[coil].values
        if method == 'idw':
            z = idw(x, y, xknown, yknown, values)
            z = z.reshape(X.shape)
        elif method == 'kriging':
            from pykrige.ok import OrdinaryKriging
            gridx = np.linspace(xmin, xmax, nx)
            gridy = np.linspace(ymin, ymax, ny)
            OK = OrdinaryKriging(xknown, yknown, values, variogram_model='linear',
                                 verbose=True, enable_plotting=False, nlags=25)
            z, ss = OK.execute('grid', gridx, gridy)
        else:
            z = griddata(np.c_[xknown, yknown], values, (X, Y), method=method)
        inside = np.ones(nx*ny)
        inside2 = clipConvexHull(xknown, yknown, x, y, inside)
        ie = np.isnan(inside2).reshape(z.shape)
        z[ie] = np.nan
        Z = np.flipud(z.T)
        
        # distance between corners
        dist0 = np.abs(xmax - xmin)
        dist1 = np.abs(ymax - ymin)        
    
        Z = np.fliplr(np.flipud(Z.T))
        yscale = dist1/Z.shape[0]
        xscale = dist0/Z.shape[1]
        
        tOffsetScaling = from_origin(xmin - xscale/2, ymax - yscale/2, xscale, yscale)
        tt = tOffsetScaling
        
        if color == True:
            if vmin is None:
                vmin = np.nanpercentile(Z.flatten(), 2)
            if vmax is None:
                vmax = np.nanpercentile(Z.flatten(), 98)
            norm = plt.Normalize(vmin=vmin, vmax=vmax)
            Z = plt.get_cmap(cmap, nlevel)(norm(Z))
            Z = 255*Z
            Z = Z.astype('uint8')
            for i in range(4):
                Z[np.fliplr(ie.T).T, i] = 0
        
            with rasterio.open(fname, 'w',
                           driver='GTiff',
                           height=Z.shape[0],
                           width=Z.shape[1], count=4, dtype=Z.dtype,
                           crs='+init=epsg:27700', transform=tt) as dst:
                for i in range(4):
                    dst.write(Z[:,:,i], i+1)
        else:
            with rasterio.open(fname, 'w',
                               driver='GTiff',
                               height=Z.shape[0],
                               width=Z.shape[1], count=1, dtype=Z.dtype,
                               crs='+init=epsg:27700', transform=tt) as dst:
                dst.write(Z, 1)
                

    def pointsKiller(self):
        """Interactively kill points. Then save df after that.
        """
        pass
        
        geom = True
        resist = self.df['resist'].values
        clabel = 'Apparent Resistivity [$\Omega.m$]'
        if label == '':
            label = clabel
        inan = np.isnan(resist)
        resist = resist.copy()[~inan]
        array = array.copy()[~inan]
        self.iselect = np.zeros(len(inan), dtype=bool)
        
        def setSelect(ie, boolVal):
            ipoints[ie] = boolVal
            self.iselect[~inan] = ipoints
        spacing = np.mean(np.diff(self.elec[:,0]))
        nelec = np.max(array)
        elecpos = np.arange(0, spacing*nelec, spacing)
        
        self.eselect = np.zeros(len(elecpos), dtype=bool)
        
        if geom: # compute and applied geometric factor
            apos = elecpos[array[:,0]-1]
            bpos = elecpos[array[:,1]-1]
            mpos = elecpos[array[:,2]-1]
            npos = elecpos[array[:,3]-1]
            AM = np.abs(apos-mpos)
            BM = np.abs(bpos-mpos)
            AN = np.abs(apos-npos)
            BN = np.abs(bpos-npos)
            K = 2*np.pi/((1/AM)-(1/BM)-(1/AN)+(1/BN)) # geometric factor
            resist = resist*K
            
        if log:
            resist = np.sign(resist)*np.log10(np.abs(resist))
        
        array = np.sort(array, axis=1) # need to sort the array to make good wenner pseudo section
        cmiddle = np.min([elecpos[array[:,0]-1], elecpos[array[:,1]-1]], axis=0) \
            + np.abs(elecpos[array[:,0]-1]-elecpos[array[:,1]-1])/2
        pmiddle = np.min([elecpos[array[:,2]-1], elecpos[array[:,3]-1]], axis=0) \
            + np.abs(elecpos[array[:,2]-1]-elecpos[array[:,3]-1])/2
        xpos = np.min([cmiddle, pmiddle], axis=0) + np.abs(cmiddle-pmiddle)/2
        ypos = - np.sqrt(2)/2*np.abs(cmiddle-pmiddle)
        
        
        def onpick(event):
            if lines[event.artist] == 'data':
                xid, yid = xpos[event.ind[0]], ypos[event.ind[0]]
                isame = (xpos == xid) & (ypos == yid)
                if (ipoints[isame] == True).all():
                    setSelect(isame, False)
                else:
                    setSelect(isame, True)
            
            if lines[event.artist] == 'elec':
                ie = (array == (event.ind[0]+1)).any(-1)
                if all(ipoints[ie] == True):
                    setSelect(ie, False)
                else:
                    setSelect(ie, True)
                if self.eselect[event.ind[0]] == True:
                    self.eselect[event.ind[0]] = False
                else:
                    self.eselect[event.ind[0]] = True
                elecKilled.set_xdata(elecpos[self.eselect])
                elecKilled.set_ydata(np.zeros(len(elecpos))[self.eselect])
            killed.set_xdata(x[ipoints])
            killed.set_ydata(y[ipoints])
            killed.figure.canvas.draw()                
                
        if ax is None:
            fig, ax = plt.subplots()
        else:
            fig = ax.figure
        caxElec, = ax.plot(elecpos, np.zeros(len(elecpos)), 'ko', picker=5)
        cax = ax.scatter(xpos, ypos, c=resist, marker='o', picker=5, vmin=vmin,
                         vmax=vmax)
        cbar = fig.colorbar(cax, ax=ax)
        cbar.set_label(label)
        cax.figure.canvas.mpl_connect('pick_event', onpick)
        
        killed, = cax.axes.plot([],[],'rx')
        elecKilled, = cax.axes.plot([],[],'rx')
        x = cax.get_offsets()[:,0]
        y = cax.get_offsets()[:,1]
        
        ipoints = np.zeros(len(y),dtype=bool)

        lines = {cax:'data',caxElec:'elec',killed:'killed'}
          
    
    
    
    
    def gridData(self, nx=100, ny=100, method='nearest', xmin=None, xmax=None,
                 ymin=None, ymax=None):
        """ Grid data (for 3D).
        
        Parameters
        ----------
        nx : int, optional
            Number of points in x direction.
        ny : int, optional
            Number of points in y direction.
        xmin : float, optional
            Mininum X value.
        xmax : float, optional
            Maximum X value.
        ymin : float, optional
            Minimium Y value.
        ymax : float, optional
            Maximum Y value
        method : str, optional
            Interpolation method (nearest, cubic or linear see
            `scipy.interpolate.griddata`) or IDW (default).
        """
        x = self.df['x'].values
        y = self.df['y'].values
        if xmin is None:
            xmin = np.min(x)
        if xmax is None:
            xmax = np.max(x)
        if ymin is None:
            ymin = np.min(y)
        if ymax is None:
            ymax = np.max(y)
        X, Y = np.meshgrid(np.linspace(xmin, xmax, nx),
                           np.linspace(ymin, ymax, ny))
        inside = np.ones(nx*ny)
        inside2 = clipConvexHull(self.df['x'].values,
                                 self.df['y'].values,
                                 X.flatten(), Y.flatten(), inside)
        ie = ~np.isnan(inside2)
        df = pd.DataFrame()
        df['x'] = X.flatten()
        df['y'] = Y.flatten()
        for coil in self.coils:
            values = self.df[coil].values
            if method == 'idw':
                z = idw(X.flatten(), Y.flatten(), x, y, values)
            else:
                z = griddata(np.c_[x, y], values, (X, Y), method=method)
            df[coil] = z.flatten()
        self.dfg = df[ie]
        # TODO add OK kriging ?
        
    
    def crossOverPoints(self, coil=None, ax=None, dump=print, minDist=1):
        """ Build an error model based on the cross-over points.
        
        Parameters
        ----------
        coil : str, optional
            Name of the coil.
        ax : Matplotlib.Axes, optional
            Matplotlib axis on which the plot is plotted against if specified.
        dump : function, optional
            Output function for information.
        minDist : float, optional
            Point at less than `minDist` from each other are considered
            identical (cross-over). Default is 1 meter.
        """
        if coil is None:
            coil = self.coils[0]
        df = self.df
        dist = cdist(df[['x', 'y']].values,
                     df[['x', 'y']].values)
        ix, iy = np.where(((dist < minDist) & (dist > 0))) # 0 == same point
        ifar = (ix - iy) > 200 # they should be at least 200 measuremens apart
        ix, iy = ix[ifar], iy[ifar]
        print('found', len(ix), '/', df.shape[0], 'crossing points')
        
        if len(ix) < 10:
            dump('None or too few colocated measurements found for error model.')
            return
        
        val = df[coil].values
        x = val[ix]
        y = val[iy]
        means = np.mean(np.c_[x,y], axis=1)
        error = np.abs(x - y)
        
        # bin data (constant number)
        nbins = 30 # number of data per bin
        end = int(np.floor(len(means)/nbins)*nbins)
        errorBinned = error[:end].reshape((-1, nbins)).mean(axis=1)
        meansBinned = means[:end].reshape((-1, nbins)).mean(axis=1)
        
        # bin data (constant width)
        # errorBinned, binEdges, _ = binned_statistic(
        #         means, error, 'mean', bins=20)
        # meansBinned = binEdges[:-1] + np.diff(binEdges)
        

        # compute model
        inan = ~np.isnan(meansBinned) & ~np.isnan(errorBinned)
        inan = inan & (meansBinned > 0) & (errorBinned > 0)
        slope, intercept, r_value, p_value, std_err = linregress(
                np.log10(meansBinned[inan]), np.log10(errorBinned[inan]))
        
        self.df[coil + '_err'] = intercept + slope * self.df[coil]
            
        # plot
        if ax is None:
            fig, ax = plt.subplots()
        ax.set_title(coil)
        ax.loglog(means, error, '.')
        ax.loglog(meansBinned, errorBinned, 'o')
        predError = 10**(intercept + slope * np.log10(means))
        eq = r'$\epsilon = {:.2f} \times \sigma^{{{:.2f}}}$'.format(10**intercept, slope)
        isort = np.argsort(means)
        ax.loglog(means[isort], predError[isort], 'k-', label=eq)
        ax.legend()
        ax.set_xlabel(r'Mean $\sigma_a$ [mS/m]')
        ax.set_ylabel(r'Error $\epsilon$ [mS/m]')
     
    
    
    def plotCrossOverMap(self, coil=None, ax=None, minDist=1):
        """Plot the map of the cross-over points for error model.
        
        Parameters
        ----------
        coil : str, optional
            Name of the coil.
        ax : Matplotlib.Axes, optional
            Matplotlib axis on which the plot is plotted against if specified.
        minDist : float, optional
            Point at less than `minDist` from each other are considered
            identical (cross-over). Default is 1 meter.
        """
        if coil is None:
            coil = self.coils[0]
        df = self.df
        dist = cdist(df[['x', 'y']].values,
                     df[['x', 'y']].values)
        ix, iy = np.where(((dist < minDist) & (dist > 0))) # 0 == same point
        ifar = (ix - iy) > 200 # they should be at least 200 measuremens apart
        ix, iy = ix[ifar], iy[ifar]
        print('found', len(ix), '/', df.shape[0], 'crossing points')
        
        # plot cross-over points
        xcoord = df['x'].values
        ycoord = df['y'].values
        icross = np.unique(np.r_[ix, iy])
        
        if ax is None:
            fig1, ax = plt.subplots()
        ax.set_title(coil)
        ax.plot(xcoord, ycoord, '.')
        ax.plot(xcoord[icross], ycoord[icross], 'ro', label='crossing points')
        ax.set_xlabel('x [m]')
        ax.set_ylabel('y [m]')

    
    def gfCorrection(self):
        """ Apply the correction due to the 1m calibration.
        """
        pass
        # TODO
        
    def importGF(self, fnameLo=None, fnameHi=None, device='CMD Mini-Explorer',
                 hx=0, targetProjection='EPSG:27700'):
        """Import GF instrument data with Lo and Hi file mode. If spatial data
        a regridding will be performed to match the data.
        
        Parameters
        ----------
        fnameLo : str
            Name of the file with the Lo settings.
        fnameHi : str, optional
            Name of the file with the Hi settings.
        device : str, optional
            Type of device. Default is Mini-Explorer.
        hx : float, optional
            Height of the device above the ground in meters according to the
            calibration use (e.g. `F-Ground` -> 0 m, `F-1m` -> 1 m).
        targetProjection : str, optional
            If both Lo and Hi dataframe contains 'Latitude' with NMEA values
            a conversion first is done using `self.convertFromNMEA()` before
            being regrid using nearest neightbours.
        """
        if fnameLo is None and fnameHi is None:
            raise ValueError('You must specify at least one of fnameLo or fnameHi.')
        if fnameLo is not None:
            self.name = os.path.basename(fnameLo)[:-4] # to remove .dat
        else:
            self.name = os.path.basename(fnameHi)[:-4]
            
        if device == 'CMD Mini-Explorer':
            freq = 30000
            csep = [0.32, 0.71, 1.18]
        elif device == 'CMD Explorer':
            freq = 10000
            csep = [1.48, 2.82, 4.49]
        else:
            raise ValueError('Device ' + device + ' unknown.')

        loCols = ['VCP{:.2f}'.format(a) for a in csep]
        loCols += [a + '_inph' for a in loCols]
        hiCols = ['HCP{:.2f}'.format(a) for a in csep]
        hiCols += [a + '_inph' for a in hiCols]
        cols = ['Cond.1[mS/m]', 'Cond.2[mS/m]', 'Cond.3[mS/m]',
                'Inph.1[ppt]', 'Inph.2[ppt]', 'Inph.3[ppt]']
        
        def harmonizeHeaders(df):
            x = df.columns.values
            tmp = []
            for a in x:
                tmp.append(a.replace(' [','[') # when dowloaded from usb
                            .replace('Cond1.','Cond.1') # when downloaded from usb and manual
                            .replace('Cond2.','Cond.2')
                            .replace('Cond3.','Cond.3'))
            df = df.rename(columns=dict(zip(x, tmp)))
            return df
        
        if fnameLo is not None:
            loFile = pd.read_csv(fnameLo, sep='\t')
            loFile = harmonizeHeaders(loFile)
            loFile = loFile.rename(columns=dict(zip(cols, loCols)))
        if fnameHi is not None:
            hiFile = pd.read_csv(fnameHi, sep='\t')
            hiFile = harmonizeHeaders(hiFile)
            hiFile = hiFile.rename(columns=dict(zip(cols, hiCols)))

        if fnameLo is not None and fnameHi is not None:
            if 'Latitude' not in loFile.columns and 'Latitude' not in hiFile.columns:
                if loFile.shape[0] == hiFile.shape[0]:
                    print('importGF: joining on rows.')
                    df = loFile[loCols].join(hiFile[hiCols])
                    df['x'] = np.arange(df.shape[0])
                    df['y'] = 0
                else:
                    df = None
                    raise ValueError('Can not join the dataframe as they have different lengths: {:d} and {:d}'.format(loFile.shape[0], hiFile.shape[0]))
            else:
                print('Using nearest neighbours to assign values to all merged positions.')
                # transformation of NMEA to x, y coordinates
                loFile = convertFromCoord(loFile, targetProjection)
                hiFile = convertFromCoord(hiFile, targetProjection)
                
                # regridding using nearest neighbour
                df = pd.concat([loFile, hiFile], sort=False).reset_index(drop=True)
                ie = np.zeros(df.shape[0], dtype=bool)
                ie[:loFile.shape[0]] = True
                pointsLo = loFile[['x','y']].values
                pointsHi = hiFile[['x','y']].values
                for col in loCols:
                    values = loFile[col].values
                    interpolator = NearestNDInterpolator(pointsLo, values)
                    df.loc[~ie, col] = interpolator(pointsHi)
                for col in hiCols:
                    values = hiFile[col].values
                    interpolator = NearestNDInterpolator(pointsHi, values)
                    df.loc[ie, col] = interpolator(pointsLo)
                    
            coils = loCols[:3] + hiCols[:3]
            coilsInph = loCols[3:] + hiCols[3:]
        elif fnameLo is not None:
            df = loFile
            df['x'] = np.arange(df.shape[0])
            df['y'] = 0
            coils = loCols[:3]
            coilsInph = loCols[3:]
        elif fnameHi is not None:
            df = hiFile
            df['x'] = np.arange(df.shape[0])
            df['y'] = 0
            coils = hiCols[:3]
            coilsInph = hiCols[3:]

        if df is not None:
            self.coils = coils
            self.coilsInph = coilsInph
            coilInfo = [self.getCoilInfo(c) for c in self.coils]
            self.freqs = np.repeat([freq], len(self.coils))
            self.cspacing = [a['coilSeparation'] for a in coilInfo]
            self.cpos = [a['orientation'] for a in coilInfo]
            self.hx = np.repeat([hx], len(self.coils))*0 # as we corrected it before
            # applying correction for GF instruments
            print('Transformation to LIN ECa for {:s} at F-{:.0f}m calibration'.format(device, hx))
            if hx == 0:
                gfcoefs = {'HCP1.48': 24.87076856,
                           'HCP2.82': 7.34836983,
                           'HCP4.49': 3.18322873,
                           'VCP1.48': 23.96851467,
                           'VCP2.82': 6.82559412,
                           'VCP4.49': 2.81033124,
                           'HCP0.32': 169.35849385,
                           'HCP0.71': 35.57031603,
                           'HCP1.18': 13.42529324,
                           'VCP0.32': 167.10523718,
                           'VCP0.71': 34.50404729,
                           'VCP1.18': 12.744378}
                for i, coil in enumerate(coils):
                    qvalues = 0+df[coil].values/gfcoefs[coil]*1e-3j
                    df.loc[:, coil] = Q2eca(qvalues, self.cspacing[i], f=self.freqs[i])*1000 # mS/m
            if hx == 1:
                gfcoefs = {'HCP1.48': 43.714823,
                           'HCP2.82': 9.22334343,
                           'HCP4.49': 3.51201955,
                           'VCP1.48': 77.90907085,
                           'VCP2.82': 14.02757873,
                           'VCP4.49': 4.57001088}
                ''' from mS/m to Q in ppt using GF calibration
                from Q (not in ppt) to ECa LIN using Q2eca
                '''
                for i, coil in enumerate(coils):
                    qvalues = 0+df[coil].values/gfcoefs[coil]*1e-3j # in part per thousand
                    df.loc[:, coil] = Q2eca(qvalues, self.cspacing[i], f=self.freqs[i])*1000
#                coefs = np.zeros(len(coils))
#                for i in range(len(coils)):
#                    coefs[i] = emSens(np.array([1]), self.cspacing[i], self.cpos[i], hx=0)[0]
#                df.loc[:,coils] = df.loc[:, coils]*coefs
            self.df = df
            self.sensor = device
        
        
    ### jamyd91 contribution ### 
    def consPtStat(self):# work out distance between consective points 
        """compute geometrical statistics of consective points, azimuth and 
        bearing of walking direction, time between consective measurements and
        distance between consecutive measurements. Results appended to self.df. 
        
        Notes
        ---------
        Requires local coordinates system to be assigned first! 
        """
        df = self.df
        try:
            x = df.x.values # x values of data frame
            y = df.y.values # y values of data frame 
        except AttributeError:
            raise KeyError(" %s \n ... It looks like no local coordinate system has been assigned, try running self.convertFromNMEA() first")
        dist = np.zeros_like(x,dtype=float) # allocate array to store distance between points 
        bearing = np.zeros_like(x,dtype=float) # allocate array to store survey bearing 
        azimuth = np.zeros_like(x,dtype=float) # allocate array to store survey bearing 
        dist[0] = 999 # first value cant have a distance 
        #time handling 
        timeStr = df.Time.values # string arguments describing time 
        timeLs = timeStr[0].split(':') # parse the time string
        second = float(timeLs[2]) # seconds
        micro = (second - int(second))*100000 # parse the micro second argument 
        times = [time(hour = int(timeLs[0]), #make python datetime.time class
                           minute = int(timeLs[1]), 
                           second = int(second),
                           microsecond=int(micro)
                           )]*len(x)
        current_date = datetime.now()
        current_year = current_date.year
        current_month = current_date.month
        current_day = current_date.day
        dttimes = [datetime(year=current_year,
                          month= current_month,
                          day=current_day,
                          hour = int(timeLs[0]), 
                          minute = int(timeLs[1]), 
                          second = int(second), 
                          microsecond=int(micro))]*len(x)
            
        def quadrantCheck(dx,dy):
            quad = 0
            edge = False
            if dx>0 and dy>0:#both positive
                quad = 1 # 'NE'
            elif dx>0 and dy<0:
                quad = 2 # 'SE'
            elif dx<0 and dy<0:#both negative
                quad = 3 # 'SW'
            elif dx<0 and dy>0:
                quad = 4 # 'NE'
            else:#edge case
                edge = True
                if dx==0 and dy==0:
                    quad = 0 #'0'
                elif dx==0 and dy>0:
                    quad = 0
                elif dx>0 and dy==0:
                    quad = 90
                elif dx==0 and dy<0:
                    quad = 180
                elif dx<0 and dy==0:
                    quad = 270
            return quad, edge
                    
        #go through each datframe entry and work out relevant stats 
        for i in range(1,len(x)):
            dx = x[i] - x[i-1] # delta x 
            dy = y[i] - y[i-1] # delta y 
            h = np.sqrt(dx**2 + dy**2) # hypoternues length 
            dist[i] = h
            
            #bearing stat
            quad, edge = quadrantCheck(dx,dy)
            if edge:
                az = quad
            else:
                angle = np.rad2deg(np.arcsin(dx/h)) # angle of measurement direction relative to north 
                if quad == 1:
                    az = angle
                elif quad == 2:
                    az = 180 - angle 
                elif quad == 3:
                    az = 180 + abs(angle)
                elif quad == 4:
                    az = 360 - abs(angle) 
            
            azimuth[i] = az
            if az > 180: #if over 180
                bearing[i] = az - 180 # add 180 in order to get a postive bearing or strike 
            else:
                bearing[i] = az
            
            #time related stats
            timeLs = timeStr[i].split(':') # parse the time
            second = float(timeLs[2])
            micro = (second - int(second))*100000
            times[i] = time(hour = int(timeLs[0]), 
                             minute = int(timeLs[1]), 
                             second = int(second), 
                             microsecond=int(micro))
            dttimes[i] = datetime(year=current_year,#datetimes needed to get time delta objects 
                                  month= current_month,
                                  day=current_day,
                                  hour = int(timeLs[0]), 
                                  minute = int(timeLs[1]), 
                                  second = int(second), 
                                  microsecond=int(micro))
        df['conDist'] = dist # distance between consecutive points 
        df['azimuth'] = azimuth # walking direction in terms of azimuth relative to local coordinate system
        df['bearing'] = bearing # walking direction in terms of bearing relative to local coordinate system
        df['PythonTime'] = times # times in  python datetime format 
        df['elasped(sec)'] = [(dttimes[i] - dttimes[0]).seconds for i in range(len(x))]# number of seconds elasped 
            
        #return df 
        self.df = df
    
    def rmRepeatPt(self, tolerance=0.2):
        """Remove points taken too close together consecutively.
        
        Parameters
        ----------
        tolerance : float, optional
            Minimum distance away previous point in order to be retained. 
        
        Returns 
        -------
        self.fil_df : pandas dataframe
            Truncated dataframe leaving only the measurements which are spaced 
            more than [tolerance value] apart. 
        """
        #error checking
        if not isinstance(tolerance,int) and not isinstance(tolerance,float):
            raise ValueError("tolerance instance should be int or float")
        df = self.df
        try:
            vals = df.conDist
        except AttributeError as e:
            raise KeyError("Error: %s \n... Try running self.consPtStat first"%e)
        ioi = vals>tolerance
        out = df[ioi].copy()
        #return out.reset_index() 
        self.fil_df = out.reset_index()
        
    
    def rmBearing(self, phiMin, phiMax):
        """Remove measurments recorded in a certian bearing range. Where phiMax -
        phiMin is the bearing range to remove. 
        
        Parameters
        ----------
        phiMin : float, optional
            Minimum angle, in degrees. 
        phiMax : float, optional
            Maximum angle, in degrees.
        
        Returns 
        -------
        self.fil_df : pandas dataframe
            Truncated dataframe leaving only the measurements from outside the 
            given bearing range. 
        """
        #error checking
        if not isinstance(phiMin,int) and not isinstance(phiMin,float):
            raise ValueError("phiMin instance should be int or float")
        if not isinstance(phiMax,int) and not isinstance(phiMax,float):
            raise ValueError("phiMax instance should be int or float")
        if phiMin >= phiMax:
            raise ValueError("Min and max bearings cannot be the same, and min must be smaller!")
        df = self.df
        try:
            vals = df.bearing
        except AttributeError as e:
            raise KeyError("Error: %s \n... Try running self.consPtStat first"%e)
        
        ioi= [True]*len(vals)
        for i in range(len(vals)):
            if vals[i] > phiMin and vals[i] < phiMax:
                ioi[i] = False
        
        out = df[ioi].copy()
        #return out.reset_index() 
        self.fil_df = out.reset_index()# its necassary to reset the indexes for other filtering techniques 
        
        
    def driftStn(self, xStn=None, yStn=None, tolerance=0.5):
        """Extract values taken at a given x y point. By default the drift 
        station is taken at the location where the survey starts.
        
        Parameters
        ----------
        xStn : float, optional
            X coordinate of drift station (using local coordinate system, 
            not long and lat)
        yStn : float, optional
            Y coordinate of drift station (using local coordinate system, 
            not long and lat)
        tolerance : float, optional
            Maximum distance away from the drift station using local units. 
        
        Returns
        -------
        self.drift_df : pandas.DataFrame
            Truncated dataframe leaving only the measurements from the drift
            station. 
        """
        df = self.df
        x = df.x.values # x values of data frame
        y = df.y.values # y values of data frame 
        if xStn is None or yStn is None:
            xStn = x[0] # take first xy measurement as drift station
            yStn = y[0]
        
        dx = x - xStn # vector calculation
        dy = y - yStn
        dist = np.sqrt(dx**2 + dy**2) 
        ioi = dist < tolerance # index of interest 

        self.drift_df = df[ioi].copy() #return df[ioi].copy()
        
        
    def plotDrift(self, coil=None, ax=None, fit=True):
        """ Plot drift through time.
        
        Parameters
        ----------
        coil : str, optional
            Coil for which to plot the drift.
        ax : matplotlib.Axes, optional
            If specified, the graph will be plotted against.
        fit : boolean, optional
            If `True` a relatinship will be fitted.
        """
        if ax is None:
            fig, ax = plt.subplots()
        try:
            df = self.drift_df.copy()
        except AttributeError:
            self.driftStn()
            df = self.drift_df.copy()
        if coil is None:
            coil = 'Cond.1 [mS/m]'
        vals = df[coil].values
        ax.scatter(df['PythonTime'].values,vals)
        ax.set_xlabel('Time')
        ax.set_ylabel('Conductivity (mS/m)')
        if fit:
            self.fitDrift(coil=coil)
            seconds = self.drift_df['elasped(sec)'].values
            times = self.drift_df['PythonTime'].values
            cond_mdl = np.polyval(self.drift_mdl,seconds)
            ax.plot(times,cond_mdl)
        
        
    def fitDrift(self, coil=None, order=1):
        """Fit a polynomial model to the drift.
        
        Parameters
        ----------
        coil : str, optional
            Coil for which to plot the drift.
        order : int, optional
            Order of the polyfit. Default is 1.
        """
        seconds = self.drift_df['elasped(sec)'].values
        cond = self.drift_df[coil].values
        mdl = np.polyfit(seconds,cond,order)
        self.drift_mdl = mdl
        
        
    def applyDriftCorrection(self, coil=None):
        """Apply a drift correction to the coil values.
        
        Parameters
        ----------
        coil : str, optional
            Coil for which to plot the drift.
        """
        try:
            mdl = self.drift_mdl
        except AttributeError:
            self.fitDrift(coil=coil)
            mdl = self.drift_mdl
        mdl[-1] = 0 # in this case the c value should be 0 so that the model is normalised to zero
        df = self.df
        vals = df[coil].values
        correction = np.polyval(mdl,df['elasped(sec)'].values)
        self.df[coil] = vals - correction
        
        
#    def aoi(self, polyX, polyY):
#        """Identify area of interest inside a polygon.
#        
#        Parameters
#        ----------
#        polyX : ???
#        
#        polyY : ???
#        """
#        df = self.df
#        try:
#            x = df.x.values # x values of data frame
#            y = df.y.values # y values of data frame
#        except AttributeError:
#            raise KeyError(" %s \n ... It looks like no local coordinate system has been assigned, try running self.convertFromNMEA() first")        
#        inside = iip.isinpolygon(x,y,[polyX,polyY]) # dependencies not satisfied
#        self.df['AOI'] = inside


#%% test
        
if __name__ == '__main__':
#    s = Survey('test/coverCrop.csv')
    #s.show(coils='HCP0.32')
#    s.showMap(contour=True, vmax=40, pts=True)
    
#    s = Survey('test/trimpLo.csv')
#    s.convertFromNMEA()
#    s.show(s.coils[1])
#    s.keepBetween(-5,11)
#    s.rollingMean(10)
#    s.show()
    
#    s.convertFromNMEA()
#    s.showMap(contour=True, pts=True)
#    s.crossOverPoints(s.coils[1])
#    s.gridData(method='idw')
#    s.showMap(s.coils[1])
#    s.df = s.dfg
#    s.showMap(s.coils[1], contour=True)



#%%
    s = Survey('emagpy/test/coverCrop.csv')
#    s.importGF('test/potatoesLo.dat', 'test/potatoesHi.dat', hx=0)
#    s.importGF('test/trimpLo.dat', 'test/trimpHi.dat', hx=1, device='CMD Explorer')
#    s.readFile('test/potatoesLo.csv')
#    s.convertFromNMEA()
#    s.consPtStat() # bearing
#    s.rmRepeatPt() # 
#    print(s.df.head())
#    s.showMap()
#    s.saveMap('test/potatoes.tiff', nx=100, ny=300)
    s.show(s.coils[0])


# to build run : python3 setup.py sdist bdist_wheel
import setuptools
from emagpy import EMagPy_version


with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="emagpy",
    version=EMagPy_version,
#    version="1.1.11",
    author="HKEx",
    description="API for FDEM inversion and data manipulation",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://gitlab.com/hkex/emagpy",
    packages=setuptools.find_packages(),
    install_requires=['numpy','matplotlib','pandas','scipy', 'hankel', 'pyproj'],
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
)
